from collections import Counter

from wxpy import *


class TestChats:
    def test_search(self, group, friend):
        found = group.search('wxpy 机器人')
        assert friend in found
        assert isinstance(found, Chats)

    def test_stats(self, group):
        stats = group.members.stats()
        assert isinstance(stats, dict)
        for attr in 'province', 'city', 'sex':
            assert attr in stats
            assert isinstance(stats[attr], Counter)

    def test_stats_text(self, group):
        text = group.members.stats_text()
        assert '位群成员' in text



# coding: utf-8

"""
部分用例需要与 "wxpy 机器人" 进行互动
"""

import os
import time
from functools import partial
from queue import Queue

import pytest

from wxpy import *

# Todo: 修改 test，适应 new-core

_base_dir = os.path.dirname(os.path.realpath(__file__))

print('logging in test bot...')
_bot = Bot(os.path.join(_base_dir, 'wxpy_bot.pkl'))
_friend = ensure_one(_bot.friends().search('wxpy 机器人'))
_group = ensure_one(_bot.groups().search('wxpy test'))
_member = ensure_one(_group.search('游否'))

_shared_dict = dict()

attachments_dir = os.path.join(_base_dir, 'attachments')
gen_attachment_path = partial(os.path.join, attachments_dir)

global_use = partial(pytest.fixture, scope='session', autouse=True)


@global_use()
def base_dir():
    return _base_dir


@global_use()
def bot():
    return _bot


@global_use()
def friend():
    yield _friend
    while True:
        try:
            _friend.set_remark_name('')
        except ResponseError as e:
            if e.err_code == 1205:
                time.sleep(10)
                continue
        else:
            break


@global_use()
def group():
    return _group


@global_use()
def shared_dict():
    return _shared_dict


@global_use()
def member():
    return _member


@global_use()
def image_path():
    return gen_attachment_path('image.png')


@global_use()
def file_path():
    return gen_attachment_path('file.txt')


@global_use()
def video_path():
    return gen_attachment_path('video.mp4')


def wait_for_message(chats=None, msg_types=None, except_self=True, timeout=30):
    """
    等待一条指定的消息，并返回这条消息

    :param chats: 所需等待消息所在的聊天会话
    :param msg_types: 所需等待的消息类型
    :param except_self: 是否排除自己发送的消息
    :param timeout: 等待的超时秒数，若为 None 则一直等待，直到收到所需的消息
    :return: 若在超时内等到了消息，则返回此消息，否则抛出 `queue.Empty` 异常
    """

    received = Queue()

    @_bot.register(chats=chats, msg_types=msg_types, except_self=except_self)
    def _func(msg):
        received.put(msg)

    _config = _bot.registered.get_config_by_func(_func)

    ret = received.get(timeout=timeout)

    _bot.registered.remove(_config)

    return ret



# coding: utf-8
from __future__ import unicode_literals

import time
from ctypes import c_int32
from wxpy.compatible import PY2

if PY2:
    from urllib import quote
    from urlparse import urlparse, urljoin
else:
    from urllib.parse import quote, urljoin, urlparse

# 以下大部分值来自
# https://res.wx.qq.com/a/wx_fed/webwx/res/static/js/index_4a48aef.js


class URIS(object):
    HOST_CASES = (
        # 可以通过正则替换从 JS 中获取
        # 查找: .+?(?<=t\.indexOf\(")(.+?)(?=").+?(?<=a = ")(.+?)(?=").+?(?<=n = ")(.+?)(?=").+?(?<=i = ")(.+?)(?=").+?
        # 替换为: ('$1', '$2', '$3', '$4'),\n
        ('wx2.qq.com', 'login.wx2.qq.com', 'file.wx2.qq.com', 'webpush.wx2.qq.com'),
        ('wx8.qq.com', 'login.wx8.qq.com', 'file.wx8.qq.com', 'webpush.wx8.qq.com'),
        ('qq.com', 'login.wx.qq.com', 'file.wx.qq.com', 'webpush.wx.qq.com'),
        ('web2.wechat.com', 'login.web2.wechat.com', 'file.web2.wechat.com', 'webpush.web2.wechat.com'),
        ('wechat.com', 'login.web.wechat.com', 'file.web.wechat.com', 'webpush.web.wechat.com'),
    )

    START_PAGE = 'https://wx.qq.com/'
    QR_DOWNLOAD = 'https://login.weixin.qq.com/qrcode/'
    QR_LOGIN = 'https://login.weixin.qq.com/l/'

    # noinspection SpellCheckingInspection
    def __init__(self, current_page=START_PAGE, language='zh_CN'):
        """
        初始化和提供 Web 微信中各接口所需的 URI
        :param current_page: Web 微信当前所处的页面 URL
        """

        self.current_page = current_page
        self.language = language

        self.host = urlparse(self.current_page).netloc
        self.path_prefix = '/cgi-bin/mmwebwx-bin'
        self.base = urljoin(self.current_page, self.path_prefix)

        a = 'login.weixin.qq.com'
        n = 'file.wx.qq.com'
        i = 'webpush.weixin.qq.com'

        for case in self.HOST_CASES:
            if case[0] in self.current_page:
                a, n, i = case[1:]
                break

        protocol = 'https://'

        self.js_login = protocol + a + '/jslogin?appid=wx782c26e4c19acffb&redirect_uri=' + quote(
            self.base + '/webwxnewloginpage') + '&fun=new&lang=' + self.language
        self.login = protocol + a + self.path_prefix + '/login'
        self.sync_check = protocol + i + self.path_prefix + '/synccheck'
        self.download_media = protocol + n + self.path_prefix + '/webwxgetmedia'
        self.upload_media = protocol + n + self.path_prefix + '/webwxuploadmedia'

        self.preview = self.base + '/webwxpreview'
        self.init = self.base + '/webwxinit?r={}'.format(self.ts_invert)
        self.get_contact = self.base + '/webwxgetcontact'
        self.sync = self.base + '/webwxsync'
        self.batch_get_contact = self.base + '/webwxbatchgetcontact'
        self.get_icon = self.base + '/webwxgeticon'
        self.send_msg = self.base + '/webwxsendmsg'
        self.send_msg_img = self.base + '/webwxsendmsgimg'
        self.send_video_msg = self.base + '/webwxsendvideomsg'
        self.send_emoticon = self.base + '/webwxsendemoticon'
        self.send_app_msg = self.base + '/webwxsendappmsg'
        self.get_head_img = self.base + '/webwxgetheadimg'
        self.get_msg_img = self.base + '/webwxgetmsgimg'
        self.get_media = self.base + '/webwxgetmedia'
        self.get_video = self.base + '/webwxgetvideo'
        self.logout = self.base + '/webwxlogout'
        self.get_voice = self.base + '/webwxgetvoice'
        self.update_chatroom = self.base + '/webwxupdatechatroom'
        self.create_chatroom = self.base + '/webwxcreatechatroom'
        self.status_notify = self.base + '/webwxstatusnotify'
        self.check_url = self.base + '/webwxcheckurl'
        self.verify_user = self.base + '/webwxverifyuser'
        self.feedback = self.base + '/webwxsendfeedback'
        self.stat_report = self.base + '/webwxstatreport'
        self.search_contact = self.base + '/webwxsearchcontact'
        self.op_log = self.base + '/webwxoplog'
        self.check_upload = self.base + '/webwxcheckupload'
        self.revoke_msg = self.base + '/webwxrevokemsg'
        self.push_login_url = self.base + '/webwxpushloginurl'

        self._ts_add_up = int(time.time() * 1000)
        self._tip_called = False

        self.upload_media_count = 0

    @property
    def ts_now(self):
        return int(time.time() * 1000)

    @property
    def ts_invert(self):
        # 模拟 JavaScript 中的 `~new Date`
        return ~ c_int32(int(time.time() * 1000)).value

    @property
    def ts_add_up(self):
        self._ts_add_up += 1
        return self._ts_add_up

    @property
    def tip(self):
        if self._tip_called:
            return 0
        else:
            self._tip_called = True
            return 1


if __name__ == '__main__':
    uris = URIS()
    print(uris.ts_invert)



# coding: utf-8
from __future__ import unicode_literals

import logging

from wxpy.api.messages import SentMessage
from wxpy.api.messages.message_types import *
from wxpy.compatible import PY2
from wxpy.compatible.utils import force_encoded_string_output

if PY2:
    # noinspection PyUnresolvedReferences
    from urlparse import urljoin
else:
    from urllib.parse import urljoin

logger = logging.getLogger(__name__)


class Chat(object):
    def __init__(self, core, _chat):

        """
        基本聊天对象
            单个用户 (:class:`User`) 和群聊 (:class:`Group`) 的基础类

        :param core: 关联的内核对象
        :param _chat: username 或 原始数据 dict
        """

        self.core = core
        self.bot = self.core.bot

        self._chat = _chat

    @property
    def puid(self):
        """
        持续有效，且稳定唯一的聊天对象/用户ID，适用于持久保存
        
        请使用 :any:`Bot.enable_puid()` 来启用 puid 属性
        
        ..  tip::
        
            | :any:`puid <Chat.puid>` 是 **wxpy 特有的聊天对象/用户ID**
            | 不同于其他 ID 属性，**puid** 可始终被获取到，且具有稳定的唯一性

        ..  attention::
        
            puid 映射数据 **不可跨机器人使用**

        """

        if self.core.puid_map:
            return self.core.puid_map.get_puid(self)
        else:
            raise TypeError('puid is not enabled, you can enable it by `bot.enable_puid()`')

    @property
    def nickname(self):
        """
        该聊天对象的昵称 (好友、群员的昵称，或群名称)
        """

        nickname = self.raw.get('NickName')
        if nickname:
            return nickname
        elif self.username == self.core.username:
            return self.core.name

    @property
    def name(self):
        """
        | 该聊天对象的友好名称
        | 即: 从 备注名称、群聊显示名称、昵称(或群名称)，username 中按序选取第一个可用的
        """
        for attr in 'remark_name', 'display_name', 'nickname', 'username':
            _name = getattr(self, attr, None)
            if _name:
                return _name

    def send(self, content, msg_type=None, media_id=None):
        """
        发送消息。默认为发送文本消息，也可指定其他消息类型

        :param content: 消息类型为 TEXT 时为消息的文本内容，其他类型时为文件路径
        :param msg_type: 消息类型，支持 TEXT, IMAGE, STICKER, VIDEO, FILE (默认为 TEXT)
        :param media_id: 文件在服务器中的唯一 ID，填写后可省略上传步骤
        :return: 已发送的消息
        :rtype: :class:`SentMessage`
        """

        return self.core.send(self, content, msg_type, media_id)

    def send_image(self, path, media_id=None):
        """
        发送图片

        :param path: 图片的文件路径
        :param media_id: 文件在服务器中的唯一 ID，填写后可省略上传步骤
        :return: 已发送的消息
        :rtype: :class:`SentMessage`
        """

        return self.send(path, IMAGE, media_id)

    def send_sticker(self, path, media_id=None):
        """
        发送表情 (类似于手机端中发送收藏的表情)

        :param path: 表情图片的文件路径
        :param media_id: 文件在服务器中的唯一 ID，填写后可省略上传步骤
        :return: 已发送的消息
        :rtype: :class:`SentMessage`
        """

        return self.send(path, STICKER, media_id)

    def send_video(self, path, media_id=None):
        """
        发送视频

        :param path: 视频的文件路径 (通常为 mp4 格式)
        :param media_id: 文件在服务器中的唯一 ID，填写后可省略上传步骤
        :return: 已发送的消息
        :rtype: :class:`SentMessage`
        """

        return self.send(path, VIDEO, media_id)

    def send_file(self, path, media_id=None):
        """
        发送文件

        :param path: 文件路径
        :param media_id: 文件在服务器中的唯一 ID，填写后可省略上传步骤
        :return: 已发送的消息
        :rtype: :class:`SentMessage`
        """

        return self.send(path, FILE, media_id)

    def send_card(self, wxid, friend_or_mp):
        """
        发送好友名片或公众号名片 (被发送的名片必须为好友关系或已关注的公众号)

        :param wxid: 好友或公众号的微信ID (无法直接获取，需要手动填写)
        :param friend_or_mp: 好友对象或公众号对象
        """

        card_name = friend_or_mp.nickname if isinstance(friend_or_mp, Chat) else friend_or_mp
        logger.info('sending {} to {}: {}'.format(CARD, self, card_name))

        return self.core.send(
            self, content='<msg username="{}" nickname="{}"/>'.format(wxid, card_name),
            msg_type=CARD, url=self.core.uris.send_msg, text=card_name
        )

    def mark_as_read(self):
        """
        消除当前聊天对象的未读提示小红点
        """

        logger.debug('marking {} as read'.format(self))

        return self.core.status_notify(self.core.username, self.username, 1)

    def pin(self):
        """
        将聊天对象置顶
        """
        logger.info('pinning {}'.format(self))
        return self.core.op_log(self, 3, 1)

    def unpin(self):
        """
        取消聊天对象的置顶状态
        """
        logger.info('unpinning {}'.format(self))
        return self.core.op_log(self, 3, 0)

    def get_avatar(self, save_path=None):
        """
        获取头像

        :param save_path: 保存路径(后缀通常为.jpg)，若空则直接返回字节数据
        """

        logger.info('getting avatar of {}'.format(self))

        if self.raw.get('HeadImgUrl'):
            return self.core.download(urljoin(self.core.uris.current_page, self.raw['HeadImgUrl']), save_path)
        else:
            logger.warning("raw info for {} is {}".format(self, self.raw))
            raise ValueError('failed to get the avatar url for {}'.format(self))

    @property
    def raw(self):
        """
        原始数据
        """

        if isinstance(self._chat, dict):
            return self._chat
        else:
            return self.core.data.raw_chats.get(self._chat, dict())

    @property
    def username(self):

        """
        该聊天对象的内部 ID，会随着登陆会话而改变，通常不需要用到

        ..  attention::
            此 ID 在机器人重新登录后 **会被改变** !
        """
        if PY2:
            if isinstance(self._chat, unicode):
                return self._chat
            else:
                return self._chat['UserName']
        else:
            if isinstance(self._chat, str):
                return self._chat
            else:
                return self._chat['UserName']

    def update(self):
        """
        更新聊天对象的详细信息

        ..  tip::

            | 对于群聊对象，`group.update()` 仅更新群本身，而不会更新群成员的详细信息
            | 若要更新群员信息，可使用 `group.members.update()`
        """

        self.core.batch_get_contact(self)

    @force_encoded_string_output
    def __repr__(self):
        return self.__unicode__()

    def __unicode__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __cmp__(self, other):
        return 0 if self.__eq__(other) else 1

    def __hash__(self):
        return hash((Chat, self.username))



# coding: utf-8
from __future__ import unicode_literals

import logging

from .chat import Chat

logger = logging.getLogger(__name__)


class User(Chat):
    """
    好友(:class:`Friend`)、群聊成员(:class:`Member`)，和公众号(:class:`MP`) 的基础类
    """

    @property
    def remark_name(self):
        """
        备注名称
        """
        return self.raw.get('RemarkName') or None

    def set_remark_name(self, remark_name):
        """
        设置或修改好友的备注名称

        :param remark_name: 新的备注名称
        """

        logger.info('setting remark name for {}: {}'.format(self, remark_name))

        return self.core.op_log(self, 2, remark_name=remark_name or '')

    @property
    def sex(self):
        """
        性别，目前有::
        
            # 男性
            MALE = 1
            # 女性
            FEMALE = 2
        
        未设置时为 `None`
        """
        return self.raw.get('Sex') or None

    @property
    def province(self):
        """
        省份
        """
        return self.raw.get('Province') or None

    @property
    def city(self):
        """
        城市
        """
        return self.raw.get('City') or None

    @property
    def signature(self):
        """
        个性签名
        """
        return self.raw.get('Signature') or None

    @property
    def is_friend(self):
        """
        判断当前用户是否为好友关系

        :return: 若为好友关系，返回对应的好友，否则返回 False
        """

        if self.username in self.core.data.raw_chats:
            return self.core.get_chat_obj(self.username)
        return False

    def add(self, verify_content=None):
        """
        添加当前用户为好友

        :param verify_content: 验证信息(文本)
        """

        from .mp import MP

        if isinstance(self, MP):
            return self.bot.add_mp(self, verify_content)
        else:
            return self.bot.add_friend(self, verify_content)

    def accept(self):
        """
        接受当前用户为好友

        :return: 新的好友对象
        :rtype: :class:`wxpy.Friend`
        """
        return self.bot.accept_friend(self)



# coding: utf-8
from __future__ import unicode_literals

import logging
import time
from functools import wraps

from wxpy.exceptions import ResponseError

logger = logging.getLogger(__name__)


def dont_raise_response_error(func):
    """
    装饰器：用于避免被装饰的函数在运行过程中抛出 ResponseError 错误
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ResponseError as e:
            logger.warning('{0.__class__.__name__}: {0}'.format(e))

    return wrapped


def ensure_one(found):
    """
    确保列表中仅有一个项，并返回这个项，否则抛出 `ValueError` 异常

    通常可用在查找聊天对象时，确保查找结果的唯一性，并直接获取唯一项

    :param found: 列表
    :return: 唯一项
    """
    if not isinstance(found, list):
        raise TypeError('expected list, {} found'.format(type(found)))
    elif not found:
        raise ValueError('not found')
    elif len(found) > 1:
        raise ValueError('more than one found: {}'.format(found))
    else:
        return found[0]


def mutual_friends(*args):
    """
    找到多个微信用户的共同好友

    :param args: 每个参数为一个微信用户的机器人(Bot)，或是聊天对象合集(Chats)
    :return: 共同好友列表
    :rtype: :class:`wxpy.Chats`
    """

    from wxpy.api.bot import Bot
    from wxpy.api.chats import Chats, User

    class FuzzyUser(User):
        def __init__(self, user):
            super(FuzzyUser, self).__init__(user.core, user.raw)

        def __hash__(self):
            return hash((self.nickname, self.sex, self.province, self.city, self.raw['AttrStatus']))

    mutual = set()

    for arg in args:
        if isinstance(arg, Bot):
            friends = map(FuzzyUser, arg.friends)
        elif isinstance(arg, Chats):
            friends = map(FuzzyUser, arg)
        else:
            raise TypeError

        if mutual:
            mutual &= set(friends)
        else:
            mutual.update(friends)

    return Chats(mutual)


def detect_freq_limit(func, *args, **kwargs):
    """
    检测各类 Web 微信操作的频率限制，获得限制次数和周期
    
    :param func: 需要执行的操作函数
    :param args: 操作函数的位置参数
    :param kwargs: 操作函数的命名参数
    :return: 限制次数, 限制周期(秒数)
    """

    start = time.time()
    count = 0

    while True:
        try:
            func(*args, **kwargs)
        except ResponseError as e:
            logger.info('freq limit reached: {} requests passed, error_info: {}'.format(count, e))
            break
        else:
            count += 1
            logger.debug('{} passed'.format(count))

    while True:
        period = time.time() - start
        try:
            func(*args, **kwargs)
        except ResponseError:
            logger.debug('blocking: {:.0f} secs'.format(period))
            time.sleep(1)
        else:
            logger.info('freq limit detected: {} requests / {:.0f} secs'.format(count, period))
            return count, period



from .message import Message
from .message_config import MessageConfig
from .message_types import *
from .messages import Messages
from .registered import Registered
from .sent_message import SentMessage



import pytest

from wxpy import *


class TestGroup:
    def test_group_attributes(self, group, friend, member):
        isinstance(group.members, Chats)
        assert friend in group
        assert member in group
        assert group.self == group.bot.self
        assert group.self in group
        assert not group.is_owner
        assert group.owner == friend

    def test_update_group(self, group):
        group.update(members_details=True)
        assert group.members[-1].sex is not None

    def test_add_members(self, group, member):
        try:
            group.add(member)
        except ResponseError as e:
            if e.err_code != 1205:
                raise e

    def test_remove_members(self, member):
        with pytest.raises(ResponseError) as e:
            member.remove()
            assert e.err_code == -66



from .console import embed, shell_entry
from .misc import decode_webwx_emoji, enhance_connection, ensure_list, get_raw_dict, get_receiver, \
    get_text_without_at_bot, get_username, match_attributes, match_name, match_text, new_local_msg_id, repr_message, \
    smart_map, start_new_thread
from .puid_map import PuidMap
from .tools import detect_freq_limit, dont_raise_response_error, ensure_one, mutual_friends



# coding: utf-8
from __future__ import unicode_literals

import atexit
import logging
import tempfile
import weakref
from pprint import pformat
from threading import Thread
from types import MethodType

from wxpy.api.chats import Friend, Group, MP
from wxpy.api.core import Core
from wxpy.api.messages import MessageConfig, Messages, Registered
from wxpy.api.messages.message_types import *
from wxpy.compatible import PY2
from wxpy.compatible.utils import force_encoded_string_output
from wxpy.utils import PuidMap, get_raw_dict, get_username, start_new_thread

try:
    import queue
except ImportError:
    # noinspection PyUnresolvedReferences,PyPep8Naming
    import Queue as queue

logger = logging.getLogger(__name__)


class Bot(object):
    """
    机器人对象，用于登陆和操作微信账号，涵盖大部分 Web 微信的功能::
    
        from wxpy import *
        bot = Bot()
        
        # 机器人账号自身
        myself = bot.self
        
        # 向文件传输助手发送消息
        bot.file_helper.send('Hello from wxpy!')
    """

    def __init__(self, cache_path=None, console_qr=None, qr_path=None, proxies=None, hooks=None):
        """
        :param cache_path:
            * 设置当前会话的缓存路径，并开启缓存功能；为 `None` (默认) 则不开启缓存功能。
            * 开启缓存后可在短时间内避免重复扫码，缓存失效时会重新要求登陆。
            * 设为 `True` 时，使用默认的缓存路径 'wxpy.pkl'。
        :param console_qr:
            * 在终端中显示登陆二维码，需要安装 pillow 模块 (`pip3 install pillow`)。
            * 可为整数(int)，表示二维码单元格的宽度，通常为 2 (当被设为 `True` 时，也将在内部当作 2)。
            * 也可为负数，表示以反色显示二维码，适用于浅底深字的命令行界面。
            * 例如: 在大部分 Linux 系统中可设为 `True` 或 2，而在 macOS Terminal 的默认白底配色中，应设为 -2。
        :param qr_path: 保存二维码的路径
        :param proxies: `requests 代理，形式为
            <http://docs.python-requests.org/en/master/user/advanced/#proxies>`_
        :param hooks: 用于快速重载 :class:`Core` 中的各种方法，形式为 `{'原方法名': 新函数, ...}`
        """

        self.core = Core(
            bot=weakref.proxy(self), cache_path=cache_path,
            console_qr=console_qr, qr_path=qr_path, proxies=proxies
        )

        if isinstance(hooks, dict):
            for ori_method_name, replace_func in hooks.items():
                # noinspection PyArgumentList
                setattr(self.core, ori_method_name, MethodType(replace_func, self.core))

        self.core_thread = self.core.login()

        self.self = self.core.self
        self.username = self.core.username

        self.messages = Messages()
        self.registered = Registered(self)

        # Todo: 重新实现 puid，以及 puid 的版本迁移
        self.puid_map = None
        # Todo: 自动标为已读: 改为以秒数间隔的操作，避免操作过于频繁
        self.auto_mark_as_read = False

        self.is_listening = False
        self.listening_thread = None

        if PY2:
            from wxpy.compatible.utils import TemporaryDirectory
            # noinspection PyArgumentList
            self.temp_dir = TemporaryDirectory(prefix='wxpy_')
        else:
            self.temp_dir = tempfile.TemporaryDirectory(prefix='wxpy_')

        self.start()

        atexit.register(self._cleanup)

    @force_encoded_string_output
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.core.name)

    def __unicode__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.core.name)

    @property
    def name(self):
        """
        机器人的微信昵称
        """

        return self.core.name

    @property
    def alive(self):
        """
        若为登陆状态，则为 True，否则为 False
        """

        return self.core_thread.is_alive()

    # chats

    def enable_puid(self, path='wxpy_puid.pkl'):
        """
        **可选操作:** 启用聊天对象的 :any:`puid <Chat.puid>` 属性::
            
            # 启用 puid 属性，并指定 puid 所需的映射数据保存/载入路径
            bot.enable_puid('wxpy_puid.pkl')
            
            # 指定一个好友
            my_friend = bot.friends().search('游否')[0]
            
            # 查看他的 puid
            print(my_friend.puid)
            # 'edfe8468'

        ..  tip::
        
            | :any:`puid <Chat.puid>` 是 **wxpy 特有的聊天对象/用户ID**
            | 不同于其他 ID 属性，**puid** 可始终被获取到，且具有稳定的唯一性

        :param path: puid 所需的映射数据保存/载入路径
        """

        self.puid_map = PuidMap(path)
        return self.puid_map

    @property
    def file_helper(self):
        """
        文件传输助手 (聊天对象)
        """
        return self.core.get_chat_obj('filehelper')

    @property
    def chats(self):
        """
        所有聊天对象

        :rtype: :class:`wxpy.Chats`
        """

        return self.core.get_chats(convert=True)

    @property
    def friends(self):
        """
        好友列表

        :rtype: :class:`wxpy.Chats`
        """

        return self.core.get_chats(Friend, convert=True)

    @property
    def groups(self):
        """
        群聊列表

        | 一些不活跃的群可能无法被获取到
        | 建议重要的群加入到通讯录中，以确保始终可见

        :rtype: :class:`wxpy.Chats`
        """

        return self.core.get_chats(Group, convert=True)

    @property
    def mps(self):
        """
        公众号列表

        :rtype: :class:`wxpy.Chats`
        """

        return self.core.get_chats(MP, convert=True)

    def get(self, keywords=None, **attributes):
        """
        查找聊天对象，等同于 :any:`Bot.chats.get(...) <Chats.get>`
        """
        return self.chats.get(keywords, **attributes)

    def find(self, keywords=None, **attributes):
        """
        查找聊天对象，等同于 :any:`Bot.chats.find(...) <Chats.find>`
        """
        return self.chats.find(keywords, **attributes)

    def search(self, keywords=None, **attributes):
        """
        查找聊天对象，等同于 :any:`Bot.chats.search(...) <Chats.search>`
        """
        return self.chats.search(keywords, **attributes)

    # add / create

    def add_friend(self, user, verify_content=''):
        """
        添加用户为好友 (注意有严格的调用频率限制!)

        :param user: 用户对象，或 username
        :param verify_content: 验证说明信息
        """

        logger.info('{}: adding {} (verify_content: {})'.format(self, user, verify_content))

        return self.core.verify_user(user, 2, verify_content)

    def add_mp(self, user):
        """
        添加/关注 公众号
        
        :param user: 公众号对象，或 username
        """

        logger.info('{}: adding {}'.format(self, user))

        return self.core.verify_user(user, 1)

    def accept_friend(self, user):
        """
        接受用户为好友

        :param user: 用户对象 (msg.card) 或 username
        :return: 新的好友对象
        :rtype: :class:`wxpy.Friend`
        """

        logger.info('{}: accepting new friend {}'.format(self, user))

        self.core.verify_user(user, 3)
        self.core.batch_get_contact(user)

        return self.friends.find(username=get_username(user)) or Friend(self.core, get_raw_dict(user))

    def create_group(self, users, topic=None):
        """
        创建一个新的群聊 (注意有严格的调用频率限制!)

        :param users: 用户列表 (不含自己，至少 2 位)
        :param topic: 群名称
        :return: 若建群成功，返回一个新的群聊对象
        :rtype: :class:`wxpy.Group`
        """

        logger.info('{}: creating group (topic: {}), with users:\n{}'.format(
            self, topic, pformat(users)))

        username = self.core.create_chatroom(users, topic)
        return self.groups.find(username=username)

    # upload

    def upload_file(self, path, msg_type):
        """
        | 上传文件，并获取 media_id
        | 可用于重复发送图片、表情、视频，和文件

        :param path: 文件路径
        :param msg_type: 发送时的消息类型，支持 IMAGE, STICKER, VIDEO, FILE
        :return: media_id
        :rtype: str
        """

        logger.info('{}: uploading file: {}'.format(self, path))

        return self.core.upload_media(path, msg_type)

    # messages / register

    def _process_message(self, msg):
        """
        处理接收到的消息
        """

        if not self.alive:
            return

        config = self.registered.get_config(msg)

        logger.debug('{}: new message (func: {}):\n{}'.format(
            self, config.func.__name__ if config else None, msg))

        if config:

            def process():
                # noinspection PyBroadException
                try:
                    ret = config.func(msg)
                    if ret is not None:
                        msg.reply(ret)
                except:
                    logger.exception('an error occurred in {}.'.format(config.func))

                if self.auto_mark_as_read and not msg.type == UNKNOWN and msg.sender != self.self:
                    from wxpy import ResponseError
                    try:
                        msg.chat.mark_as_read()
                    except ResponseError as e:
                        logger.warning('failed to mark as read: {}'.format(e))

            if config.run_async:
                start_new_thread(process, use_caller_name=True)
            else:
                process()

    def register(
            self, chats=None, msg_types=None,
            except_self=True, run_async=True, enabled=True
    ):
        """
        装饰器：用于注册消息配置

        :param chats: 消息所在的聊天对象：单个或列表形式的多个聊天对象或聊天类型，为空时匹配所有聊天对象
        :param msg_types: 消息的类型：单个或列表形式的多个消息类型，为空时匹配所有已知消息类型
        :param except_self: 排除由自己发送的消息
        :param run_async: 是否异步执行所配置的函数：可提高响应速度
        :param enabled: 当前配置的默认开启状态，可事后动态开启或关闭
        """

        def do_register(func):
            self.registered.append(MessageConfig(
                bot=self, func=func, chats=chats, msg_types=msg_types,
                except_self=except_self, run_async=run_async, enabled=enabled
            ))

            return func

        return do_register

    # noinspection PyBroadException
    def _listen(self):
        try:
            logger.info('{}: started'.format(self))
            self.is_listening = True

            while self.alive and self.is_listening:
                try:
                    msg = self.core.message_queue.get(timeout=0.5)
                except queue.Empty:
                    continue

                if msg.type != UNKNOWN:
                    self.messages.append(msg)

                try:
                    self._process_message(msg)
                except:
                    logger.exception('an error occurred while processing msg:\n{}'.format(msg))
        finally:
            self.is_listening = False
            self._cleanup()
            logger.info('{}: stopped'.format(self))

    def start(self):
        """
        开始消息监听和处理 (登陆后会自动开始)
        """

        if not self.alive:
            logger.warning('{} has been logged out!'.format(self))
        elif self.is_listening:
            logger.warning('{} is already running, no need to start again.'.format(self))
        else:
            self.listening_thread = start_new_thread(self._listen)

    def stop(self):
        """
        停止消息监听和处理 (登出后会自动停止)
        """

        if self.is_listening:
            self.is_listening = False
            self.listening_thread.join()
        else:
            logger.warning('{} is not running.'.format(self))

    def join(self):
        """
        阻塞进程，直到结束消息监听 (例如，机器人被登出时)
        """

        if isinstance(self.listening_thread, Thread):
            try:
                logger.info('{}: joined'.format(self))
                self.listening_thread.join()
            except KeyboardInterrupt:
                pass

    def logout(self):
        """
        登出当前账号
        """

        return self.core.logout()

    def _cleanup(self):
        if self.is_listening:
            self.stop()
        self.temp_dir.cleanup()

    def dump(self, force=False):
        """ 持久化 session """
        self.core.dump(force=force)



# coding: utf-8
from __future__ import unicode_literals


class ResponseError(Exception):
    """
    当 BaseResponse 的返回值不为 0 时抛出的异常
    """

    def __init__(self, core, err_code, err_msg):
        super(ResponseError, self).__init__('{}: err_code: {}; err_msg: {}'.format(core, err_code, err_msg))

        self.core = core
        self.err_code = err_code
        self.err_msg = err_msg



# coding: utf-8
from __future__ import unicode_literals

import atexit
import base64
import datetime
import hashlib
import json
import logging
import mimetypes
import os
import pickle
import platform
import random
import re
import subprocess
import time
from multiprocessing.pool import ThreadPool
from pprint import pformat
from xml.etree import ElementTree as ETree

import pyqrcode
import requests

from wxpy import __version__
from wxpy.api.chats import Chat, Chats, Friend, Group, MP, Member, User
from wxpy.api.data import Data
from wxpy.api.messages import Message, SentMessage
from wxpy.api.messages.message_types import *
from wxpy.api.uris import URIS
from wxpy.compatible import PY2
from wxpy.compatible.utils import force_encoded_string_output
from wxpy.exceptions import ResponseError
from wxpy.utils.misc import chunks, decode_webwx_json_values, diff_usernames, enhance_connection, ensure_list, \
    get_chat_type, get_username, new_local_msg_id, smart_map, start_new_thread

try:
    import queue
except ImportError:
    # noinspection PyUnresolvedReferences,PyPep8Naming
    import Queue as queue

try:
    import html
except ImportError:
    # Python 2.6-2.7
    # noinspection PyUnresolvedReferences,PyUnresolvedReferences,PyCompatibility
    from HTMLParser import HTMLParser

    html = HTMLParser()

logger = logging.getLogger(__name__)


class Core(object):
    USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_3) " \
                 "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36"

    def __init__(self, bot=None, cache_path=None, console_qr=None, qr_path=None, proxies=None):
        """
        :class:`Bot` 的内核，负责

        1. 完成初始化和登陆过程
        2. 同步检查和数据同步
        3. 包装 self.urls 中的各原始接口，供 :class:`Bot` 调用
        """

        self.bot = bot

        self.cache_path = 'wxpy.pkl' if cache_path is True else cache_path
        self.console_qr = console_qr

        if qr_path:
            self.qr_path = qr_path
        elif self.cache_path:
            # 根据 cache_path 自动生成相对应的 qr_path, 避免多开时重复
            self.qr_path = '{}_qrcode.png'.format(re.sub(r'\.\w+$', '', self.cache_path))
        else:
            self.qr_path = 'qrcode.png'

        self._proxies = proxies

        self.session = None
        self.uris = None
        self.data = Data()

        self.uuid = None
        self.qrcode = None
        self.alive = None
        self.self = None

        self.message_queue = queue.Queue()

        self.load()
        atexit.register(self.dump)

    @force_encoded_string_output
    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.name)

    @property
    def name(self):
        if self.data.raw_self:
            return self.data.raw_self.get('NickName')

    @property
    def username(self):
        return self.data.raw_self['UserName']

    # [requests]

    def new_session(self, cookies=None):

        # noinspection PyUnusedLocal
        def session_response_hook(resp, *args, **kwargs):
            resp.raise_for_status()

        self.session = requests.Session()
        self.session.headers.update({'User-Agent': self.USER_AGENT})
        self.session.hooks.update(response=session_response_hook)
        enhance_connection(self.session)

        if self._proxies:
            self.session.proxies = self._proxies

        if cookies:
            self.session.cookies = self.data.cookies
        else:
            self.uris = URIS(self.session.get(URIS.START_PAGE).url)
            self.uuid = None

        return self.session

    def get(self, url, load_resp=True, **kwargs):
        """
        经过封装的 GET 请求，会自动进行返回值检查等操作

        :param url: 请求的 URL
        :param load_resp: 是否解析 Response 中的 JSON 内容，并返回解析后的 dict
        :param kwargs: 发起请求时的其他参数
        """
        resp = self.session.get(url, **kwargs)
        if load_resp:
            resp = self.load_response_as_json(resp)
        return resp

    def post(self, url, ext_data=None, load_resp=True, **kwargs):
        """
        经过封装的 POST 请求，会自动进行返回值检查等操作

        :param url: 请求的 URL
        :param ext_data: payload 中除了 BaseRequest 字段以外的部分
        :param load_resp: 是否解析 Response 中的 JSON 内容，并返回解析后的 dict
        :param kwargs: 发起请求时的其他参数
        """

        if 'files' not in kwargs:
            data = {'BaseRequest': self.base_request}
            if ext_data:
                data.update(ext_data)
            data = json.dumps(data, ensure_ascii=False).encode('utf-8', errors='replace')
            kwargs['data'] = data
            kwargs['headers'] = {'Content-Type': 'application/json;charset=UTF-8'}

        resp = self.session.post(url, **kwargs)
        if load_resp:
            resp = self.load_response_as_json(resp)
        return resp

    def download(self, url, save_path=None):
        """
        下载文件

        :param url: 需要下载的 URL 地址
        :param save_path: 保存路径。为空时不保存到磁盘，直接返回内容数据
        """

        resp = self.session.get(url, stream=True)
        resp.raise_for_status()

        if save_path:
            with open(save_path, 'wb') as fp:
                for chunk in resp.iter_content(chunk_size=128):
                    fp.write(chunk)
                return resp
        else:
            return resp.content

    @property
    def proxies(self):
        return self._proxies

    @proxies.setter
    def proxies(self, value):
        self.session.proxies = value
        self._proxies = value

    # [apis]

    def login(self):
        """ 登陆 """

        if self.session:
            try:
                self.sync(retries=1)
            except ResponseError:
                logger.info('failed to continue last data sync loop')
            else:
                return self._logged_in()

        while True:

            if self.data.uin:
                self.uuid = self.get_push_login_uuid()
                self.data = Data()

            if not self.uuid:
                self.new_session()
                prompt('Getting uuid for QR Code')
                self.uuid = self.get_qrcode_uuid()
                # pyqrcode.QRCode 对象
                self.qrcode = pyqrcode.create(self.uris.QR_LOGIN + self.uuid)

                prompt('Downloading QR Code')
                if self.qr_path.lower().endswith('.png'):
                    qr_save_path = None
                    self.qrcode.png(self.qr_path, scale=10)
                else:
                    qr_save_path = self.qr_path
                self.download(self.uris.QR_DOWNLOAD + self.uuid, qr_save_path)

                self.show_qrcode()
                prompt('Scan the QR Code to log in')

            while True:

                # 查询 uuid 状态
                resp = self.get(
                    self.uris.login,
                    params=dict(
                        loginicon='true',
                        uuid=self.uuid,
                        tip=self.uris.tip,
                        r=self.uris.ts_invert,
                        _=self.uris.ts_add_up
                    ),
                    load_resp=False,
                )

                code = from_js(resp.text, 'window.code')
                logging.debug('got uuid status code: {}'.format(code))

                if code == 200:
                    # 验证通过

                    redirect = from_js(resp.text, 'window.redirect_uri')
                    resp = self.get(redirect, load_resp=False)

                    # 中转的 xml
                    xml_resp = resp.history[0] if resp.history else resp

                    tree = ETree.fromstring(xml_resp.text)
                    ret_code = int(tree.findtext('ret'))

                    if ret_code:
                        # 登录失败
                        logger.error('unexpected login xml content:\n{}'.format(xml_resp.text))
                        self._logged_out(ResponseError(self, ret_code, tree.findtext('message')))

                    self.data.skey = tree.findtext('skey')
                    self.data.uin = int(tree.findtext('wxuin'))
                    self.data.pass_ticket = tree.findtext('pass_ticket')
                    self.data.gray_scale = int(tree.findtext('isgrayscale'))

                    # 登陆成功
                    self.uris = URIS(resp.url)

                    prompt('Initializing')
                    self.init()

                    prompt('Loading chats data')
                    self.get_contact()
                    self.batch_get_contact(self.get_chats(Group))

                    return self._logged_in()

                elif code == 201:
                    # 需要在手机上确认登陆
                    prompt('Confirm login on your phone')
                    self.remove_qrcode()

                    avatar = from_js(resp.text, 'window.userAvatar')
                    if avatar:
                        avatar = base64.b64decode(avatar.split(',')[1])

                    self.confirm_login(avatar)
                elif code == 400:
                    # uuid 超时
                    self.uuid = None
                    prompt('UUID expired\n')
                    self.uuid_expired()
                    break
                elif code == 408:
                    # 继续等待
                    continue
                else:
                    logger.warning('unexpected uuid status:\n{}'.format(resp.text))
                    self._logged_out(ResponseError(self, code, resp.text))

    def get_push_login_uuid(self):
        """ 获取 push login 的 uuid """
        resp_json = self.get(self.uris.push_login_url, params=dict(uin=self.data.uin))
        if resp_json.get('uuid'):
            return resp_json['uuid']
        else:
            logger.warning('uuid not found while trying to push login:\n{}'.format(resp_json))

    def get_qrcode_uuid(self):
        """ 获取二维码的 uuid """
        resp = self.get(self.uris.js_login, load_resp=False)
        code, uuid = from_js(resp.text, 'window.QRLogin.code', 'window.QRLogin.uuid')
        if code == 200 and uuid:
            return uuid
        else:
            raise ResponseError(self, code, 'failed to get qrcode uuid:\n{}'.format(resp.text))

    def init(self):

        json_resp = self.post(self.uris.init)
        self.data.raw_self = json_resp['User']
        self.process_chat_list(json_resp['ContactList'])
        self.status_notify(self.data.raw_self, self.data.raw_self, 3)

    def status_notify(self, sender, receiver, code):
        return self.post(
            self.uris.status_notify,
            params=dict(lang=self.uris.language, pass_ticket=self.data.pass_ticket),
            ext_data={
                'Code': code,
                'FromUserName': get_username(sender),
                'ToUserName': get_username(receiver),
                'ClientMsgId': self.uris.ts_now,
            })

    def op_log(self, chat, cmd_id, op=None, remark_name=None):

        ext_data = {
            'CmdId': cmd_id,
            'RemarkName': remark_name or '',
            'UserName': get_username(chat),
        }

        if op:
            ext_data['OP'] = op

        return self.post(self.uris.op_log, ext_data=ext_data)

    def verify_user(self, user, op_code, verify_content=None):

        if isinstance(user, User):
            ticket = user.raw.get('Ticket', '')
        elif isinstance(user, dict):
            ticket = user.get('Ticket', '')
        else:
            ticket = ''

        return self.post(
            self.uris.verify_user,
            params=dict(r=self.uris.ts_now),
            ext_data={
                # 添加好友: 2; 通过验证: 3
                'Opcode': op_code,
                'SceneList': [33],
                'SceneListCount': 1,
                'VerifyContent': verify_content or '',
                'VerifyUserList': [{
                    'Value': get_username(user),
                    'VerifyUserTicket': ticket
                }],
                'VerifyUserListSize': 1,
                'skey': self.data.skey,
            })

    def create_chatroom(self, users, topic=None):
        """
        创建新的群聊

        :param users: 用户列表 (自己除外，至少需要两位好友)
        :param topic: 群名称
        :return: 若创建成功，返回新群的 username
        :rtype: str
        """

        usernames = list(filter(
            lambda x: x != self.username,
            map(get_username, users)
        ))

        if len(usernames) < 2:
            raise ValueError('too few users to create group')

        resp_json = self.post(
            self.uris.create_chatroom,
            params=dict(r=self.uris.ts_now),
            ext_data={
                'MemberCount': len(usernames),
                'MemberList': [{'UserName': username} for username in usernames],
                'Topic': topic or '',
            })

        if 'ChatRoomName' not in resp_json:
            raise ValueError(
                'failed to create group, '
                'missing username in the response:'
                '\n{}'.format(pformat(resp_json)))

        self.batch_get_contact(resp_json['ChatRoomName'])

        return resp_json['ChatRoomName']

    def update_chatroom(self, group, func_name, info_dict):

        """
        更新群所需的接口，例如 邀请入群、移出群员、修改群名
        """

        ext_data = {'ChatRoomName': get_username(group)}
        ext_data.update(info_dict)

        return self.post(self.uris.update_chatroom, params=dict(fun=func_name), ext_data=ext_data)

    def revoke_msg(self, receiver, server_id, local_id):
        """ 撤回消息 """
        return self.post(
            self.uris.revoke_msg,
            ext_data={
                'SvrMsgId': str(server_id),
                'ToUserName': get_username(receiver),
                'ClientMsgId': str(local_id),
            })

    def data_sync_loop(self):
        """ 主循环: 数据同步 """

        try:
            while self.alive:
                ret_code = None
                selector = None
                for _ in range(3):
                    ret_code, selector = self.sync_check()
                    if ret_code == 0:
                        break
                    elif 1100 <= ret_code <= 1102:
                        return self._logged_out(ret_code)
                    else:
                        logger.error('sync error: ret_code={}; selector={}'.format(ret_code, selector))
                else:
                    self._logged_out(ResponseError(self, ret_code, 'failed to sync'))

                if selector:
                    self.sync()
        finally:
            self.dump()

    def sync_check(self, retries=3):
        """ 检查同步状态 """
        final_error = None
        for _ in range(retries):
            resp = self.get(
                self.uris.sync_check,
                params=dict(
                    r=self.uris.ts_now,
                    skey=self.data.skey,
                    sid=self.sid,
                    uin=self.data.uin,
                    deviceid=self.device_id,
                    synckey='|'.join(['{0[Key]}_{0[Val]}'.format(d) for d in (
                            self.data.sync_check_key or self.data.sync_key)['List']]),
                    _=self.uris.ts_add_up
                ),
                timeout=(10, 30),
                load_resp=False,
            )

            logger.debug('"synccheck" resp: {}'.format(resp.text))

            try:
                found = re.search(
                    r'window\.synccheck={retcode:"(\d+)",selector:"(\d+)"}',
                    resp.text.replace(' ', '')
                )
                ret_code = int(found.group(1))
                selector = int(found.group(2))
            except (json.JSONDecodeError, TypeError, KeyError, ValueError) as e:
                logger.error('unexpected "synccheck" result:\n{}'.format(resp.text))
                final_error = e
            else:
                return ret_code, selector
        self._logged_out(final_error)

    def sync(self, retries=3, delete_when_modify=False):
        """
        同步数据

        聊天对象更新:
            新增/减少群成员: ModContactList (该群新消息时触发)
            新增/更新聊天对象: ModContactList
            减少聊天对象: DelContactList
        """

        final_error = None
        for _ in range(retries):
            try:
                resp_json = self.post(
                    self.uris.sync,
                    params=dict(
                        sid=self.sid,
                        skey=self.data.skey,
                        pass_ticket=self.data.pass_ticket
                    ),
                    ext_data={'SyncKey': self.data.sync_key, 'rr': self.uris.ts_invert},
                    timeout=(10, 30),
                )
            except ResponseError as e:
                # 在 push login 中会经历这一步，所以改用 INFO 等级
                logging.info('failed to sync: {}'.format(str(e)))
                final_error = e
            else:
                merge_chat_dict(self.data.raw_self, resp_json.get('Profile'))
                self.process_chat_list(resp_json['ModContactList'], delete=delete_when_modify)
                self.put_new_messages(resp_json['AddMsgList'])
                self.process_chat_list(resp_json['DelContactList'], delete=True)
                # ModChatRoomMemberList 在 Web 微信中未被实现
                return resp_json
        else:
            if self.alive:
                self._logged_out(final_error)
            else:
                raise final_error

    def get_contact(self):
        """ 更新通讯录列表 """
        seq = 0
        while True:
            resp_json = self.get(
                self.uris.get_contact,
                params=dict(
                    lang=self.uris.language,
                    pass_ticket=self.data.pass_ticket,
                    r=self.uris.ts_now,
                    seq=seq,
                    skey=self.data.skey,
                ))

            self.process_chat_list(resp_json['MemberList'])
            seq = resp_json.get('Seq')
            if not seq:
                break

    def batch_get_contact(self, chat_or_chats, workers=5, chunk_size=50):
        """
        批量更新聊天对象详情

        :param chat_or_chats: 需要更新的聊天对象 (单个或列表)
        :param workers: 并发线程数 (默认为 5)
        :param chunk_size: 每个请求中处理的聊天对象个数 (最高为 50)
        """

        chat_or_chats = ensure_list(chat_or_chats)

        req_list = list()
        for chat in chat_or_chats:

            if isinstance(chat, Member):
                group_username = chat.group_username
            elif isinstance(chat, dict):
                group_username = chat.get('ChatRoomId') or chat.get('EncryChatRoomId') or ''
            else:
                group_username = ''

            req_list.append({
                'UserName': get_username(chat),
                'EncryChatRoomId': group_username
            })

        def process(_chunk):
            resp_json = self.post(
                self.uris.batch_get_contact,
                params=dict(type='ex', r=self.uris.ts_now, lang=self.uris.language),
                ext_data={'Count': len(_chunk), 'List': _chunk}
            )
            self.process_chat_list(resp_json['ContactList'])

        if PY2:
            from contextlib import closing
            with closing(ThreadPool(workers)) as pool:
                pool.map(process, chunks(req_list, chunk_size))
        else:
            with ThreadPool(workers) as pool:
                pool.map(process, chunks(req_list, chunk_size))

    def send(self, receiver, content, msg_type=None, media_id=None, **kwargs):
        """
        内部使用的消息发送方法

        :param receiver: 消息的接收者
        :param content: 发送的内容。消息类型为 TEXT 时为文本，其他类型时为文件路径
        :param msg_type: 消息类型，支持 TEXT, IMAGE, STICKER, VIDEO, FILE (默认为 TEXT)
        :param media_id: 文件在服务器中的唯一 ID，填写后可省略上传步骤
        :param kwargs: 一些细节参数，目前有 url, params, file_size, text，用法详见源码
        :return: :class:`SentMessage`
        """

        msg_type = msg_type or TEXT

        logger.debug('sending {} to {}:\n{}'.format(msg_type, receiver, content))

        # url

        url = kwargs.get('url') or {
            TEXT: self.uris.send_msg,
            IMAGE: self.uris.send_msg_img,
            STICKER: self.uris.send_emoticon,
            VIDEO: self.uris.send_video_msg,
            FILE: self.uris.send_app_msg,
        }[msg_type]

        # params

        params = {'pass_ticket': self.data.pass_ticket}

        if msg_type in (IMAGE, VIDEO, FILE):
            params['fun'] = 'async'
            params['f'] = 'json'
        elif msg_type == STICKER:
            params['fun'] = 'sys'
        elif 'params' in kwargs:
            params.update(kwargs['params'])

        # msg_dict

        local_id = new_local_msg_id()
        receiver_username = get_username(receiver)
        msg_dict = {
            'ClientMsgId': local_id,
            'FromUserName': self.username,
            'LocalID': local_id,
            'ToUserName': receiver_username,
            'Type': msg_type.app or msg_type.main,
        }

        if msg_type == TEXT:
            msg_dict['Content'] = str(content)
        elif msg_type != STICKER:
            msg_dict['Content'] = ''

        if msg_type == STICKER:
            msg_dict['EmojiFlag'] = 2

        create_time = datetime.datetime.now()

        if msg_type in (IMAGE, STICKER, VIDEO, FILE):
            if not media_id:
                media_id = self.upload_media(content, msg_type, receiver)

            if msg_type == FILE:
                # noinspection SpellCheckingInspection
                msg_dict['Content'] = \
                    '<appmsg appid="wxeb7ec651dd0aefa9" sdkver=""><title>{file_name}' \
                    '</title><des/><action/><type>6</type><content/><url/><lowurl/><appattach>' \
                    '<totallen>{file_size}</totallen><attachid>{media_id}</attachid>' \
                    '<fileext>{file_ext}</fileext></appattach><extinfo/></appmsg>'.format(
                        file_name=os.path.basename(content),
                        file_size=kwargs.get('file_size') or os.path.getsize(content),
                        media_id=media_id,
                        file_ext=os.path.splitext(content)[1][1:]
                    )
            else:
                msg_dict['MediaId'] = media_id

        # request

        resp_json = self.post(url, params=params, ext_data={'Msg': msg_dict, 'Scene': 0})

        return SentMessage(
            core=self,
            type=msg_type,
            id=resp_json.get('MsgID'),
            local_id=resp_json.get('LocalID'),
            text=content if msg_type == TEXT else kwargs.get('text'),
            path=content if msg_type != TEXT else None,
            media_id=media_id,
            create_time=create_time,
            receive_time=datetime.datetime.now(),
            receiver=self.get_chat_obj(receiver_username),
        )

    def logout(self):
        """ 主动登出 """

        logger.info('logging out {}'.format(self.name))

        return self.session.post(
            self.uris.logout,
            params=dict(
                redirect=1,
                type=0,
                skey=self.data.skey
            ),
            data=dict(
                sid=self.sid,
                uin=self.data.uin
            )
        )

    @property
    def base_request(self):
        return {
            "Uin": self.data.uin, "Sid": self.sid,
            "Skey": self.data.skey, "DeviceID": self.device_id
        }

    @property
    def sid(self):
        return self.from_cookies('wxsid')

    @property
    def device_id(self):
        return 'e{}'.format(str(random.random())[2:17])

    def upload_media(self, path, msg_type, receiver=None):

        """
        上传文件，获取 media_id，可用于发送需要上传文件的消息

        *限制: 上传文件的路径中含有汉字时会导致失败*

        :param path: 文件路径
        :param msg_type: 准备用于发送的消息类型
        :param receiver: 接收者，默认为文件传输助手 (似乎并不严格)
        :return: media_id
        :rtype: str
        """

        logger.info('uploading file: {}'.format(path))

        media_type = {
            IMAGE: (1, 'pic'), STICKER: (4, 'doc'),
            VIDEO: (2, 'video'), FILE: (4, 'doc')
        }[msg_type]

        base_name = os.path.basename(path)
        file_size = os.path.getsize(path)
        file_type = mimetypes.guess_type(path)[0] or 'application/octet-stream'

        with open(path, 'rb') as fp:
            file_md5 = hashlib.md5(fp.read()).hexdigest()

        with open(path, 'rb') as fp:
            resp_json = self.post(
                self.uris.upload_media,
                params=dict(f='json'),
                files=dict(
                    id=(None, 'WU_FILE_{}'.format(self.uris.upload_media_count)),
                    name=(None, base_name),
                    type=(None, file_type),
                    lastModifiedDate=(None, time.strftime(
                        '%a %b %d %Y %H:%M:%S GMT%z (%Z)',
                        time.localtime(os.path.getmtime(path))
                    )),
                    size=(None, str(file_size)),
                    mediatype=(None, media_type[1]),
                    uploadmediarequest=(None, json.dumps({
                        'UploadType': 2,
                        'BaseRequest': self.base_request,
                        'ClientMediaId': self.uris.ts_now,
                        'TotalLen': file_size,
                        'StartPos': 0,
                        'DataLen': file_size,
                        'MediaType': media_type[0],
                        'FromUserName': self.username,
                        'ToUserName': get_username(receiver) if receiver else 'filehelper',
                        'FileMd5': file_md5,
                    })),
                    webwx_data_ticket=(None, self.from_cookies('webwx_data_ticket')),
                    pass_ticket=(None, self.data.pass_ticket),
                    filename=(base_name, fp, file_type)
                )
            )

        self.uris.upload_media_count += 1
        return resp_json['MediaId']

    # [data]
    def dump(self, force: bool = False):
        if self.cache_path:
            if not force and not self.is_living():
                logger.warning("WeChat is not living: skip dumping data!")
            else:
                logger.debug('dumping core data')
                self.data.cookies = self.session.cookies
                self.data.uris = self.uris

                with open(self.cache_path, 'wb') as fp:
                    pickle.dump(self.data, fp)

    def load(self):
        if self.cache_path and os.path.isfile(self.cache_path):
            logger.debug('loading core data')

            with open(self.cache_path, 'rb') as fp:
                data = pickle.load(fp)

            if getattr(data, 'version', None) == __version__:
                self.data = data
                self.uris = self.data.uris
                self.new_session(self.data.cookies)

    # [events]

    def show_qrcode(self):
        """ 展示二维码 """

        if not self.console_qr:
            system = platform.system()
            # noinspection PyBroadException
            try:
                if system == 'Darwin':
                    subprocess.call(['open', self.qr_path])
                elif system == 'Linux':
                    subprocess.call(['xdg-open', self.qr_path])
                else:
                    # noinspection PyUnresolvedReferences
                    os.startfile(self.qr_path)
            except:
                logger.exception('failed to open qrcode, using console_qr instead')
            else:
                return

        self.show_console_qr()

    def show_console_qr(self):
        """ 以命令行方式展示二维码 """
        # 第一种注释掉的方式在 PyCharm 的 console 中无法显示，但在其他平台中色彩更分明
        # print(qrcode.terminal(module_color=232, background=255, quiet_zone=1))
        print(self.qrcode.terminal(module_color='dark gray', background='light gray', quiet_zone=1))

    # noinspection PyMethodMayBeStatic
    def confirm_login(self, avatar):
        """
        需要在手机端确认登陆时

        :param avatar: 确认登陆时展示的头像图片数据

        hook 这个方法后，可将需要确认登陆的账号头像保存为图片::

            def save_avatar(core, avatar):
                with open('avatar.jpg', 'wb') as fp:
                    fp.write(avatar)

            bot = Bot(hooks={'confirm_login': save_avatar})
        """

        pass

    def remove_qrcode(self):
        """ 删除已扫描的二维码 """
        if os.path.isfile(self.qr_path):
            os.remove(self.qr_path)

    # noinspection PyMethodMayBeStatic
    def uuid_expired(self):
        """ 二维码 / uuid 过期时 """
        pass

    # noinspection PyMethodMayBeStatic
    def logged_in(self):
        """ 登录成功时 """
        pass

    # noinspection PyMethodMayBeStatic
    def logged_out(self, reason):
        """
        登出，或登录失败时

        :param reason:

            登出原因
            * 当自己主动登出或被服务端踢下线时为错误码 (int)
            * 当出现其他异常时，为 Exception 对象
        """
        pass

    # noinspection PyMethodMayBeStatic
    def new_message(self, msg):
        """
        收到新的消息时 (执行响应函数之前)，当需要将消息数据持久化保存时，可以 hook 该方法

        :param msg: 新的好友
        """
        pass

    # noinspection PyMethodMayBeStatic
    def new_friend(self, friend):
        """
        有新的好友时

        :param friend: 新的好友
        """
        pass

    # noinspection PyMethodMayBeStatic
    def new_group(self, group):
        """
        加入了新的群聊时

        :param group: 新的群聊对象
        """
        pass

    # noinspection PyMethodMayBeStatic
    def new_member(self, member):
        """
        有新的群成员加入时 (似乎仅在有新消息时触发)

        :param member: 新的群成员对象
        """
        pass

    # noinspection PyMethodMayBeStatic
    def deleting_friend(self, friend):
        """
        即将从本地数据中删除好友时

        :param friend: 即将被删除的好友对象
        """
        pass

    # noinspection PyMethodMayBeStatic
    def deleting_group(self, group):
        """
        即将从本地数据中删除群聊时 (似乎仅在有新消息时触发)

        :param group: 即将被删除的群聊对象
        """
        pass

    # noinspection PyMethodMayBeStatic
    def deleting_member(self, member):
        """
        即将从本地数据中删除群聊成员时

        :param member: 即将被删除的群成员对象
        """
        pass

    # [processors]

    def _logged_in(self):
        """ 这个方法仅在内部使用，请勿 hook """

        self.alive = True
        self.self = Friend(self, self.data.raw_self)

        prompt('Logged in as {}'.format(self.name))

        # 下面这个方法可以用来 hook
        self.logged_in()
        self.dump()

        # 开始并返回数据同步线程
        return start_new_thread(self.data_sync_loop)

    def _logged_out(self, reason):
        """
        这个方法仅在内部使用，请勿 hook

        :param reason: 登出原因: 可以是 错误号(int), 文字说明(str), 或异常(Exception)
        """

        self.alive = False

        # 下面这个方法可以用来 hook
        self.logged_out(reason)

        if isinstance(reason, int):
            prompt('Logged out ({})'.format(reason))
        elif isinstance(reason, str):
            prompt(reason)
        elif isinstance(reason, BaseException):
            raise reason

    def process_chat_list(self, raw_chat_list, delete=False):
        """
        处理返回数据中的聊天对象字典列表

        :param raw_chat_list: 聊天对象字典列表
        :param delete: 当为 True 时，删除聊天对象 (默认为 False)
        """

        for raw_dict in raw_chat_list:
            username = raw_dict['UserName']

            if delete:
                # 删除聊天对象
                if username in self.data.raw_chats:

                    if self.alive:
                        # 只提示登陆完成后的聊天对象变化
                        chat_obj = self.get_chat_obj(username)
                        if isinstance(chat_obj, Friend):
                            self.deleting_friend(chat_obj)
                        elif isinstance(chat_obj, Group):
                            self.deleting_group(chat_obj)

                        del self.data.raw_chats[username]

                else:
                    logger.warning('unknown chat to delete:\n{}'.format(raw_dict))

            else:
                chat_type = get_chat_type(raw_dict)
                if issubclass(chat_type, Member):
                    # 更新群成员的详细信息
                    self.data.raw_members[username] = raw_dict
                else:
                    # 新增好友、群聊，或更新群成员列表时

                    if issubclass(chat_type, Group) and raw_dict['MemberList'] and self.username not in list(map(
                            lambda x: x['UserName'], raw_dict['MemberList'])):
                        # 跳过 shadow group
                        continue

                    if self.alive:
                        # 同样的，只提示登陆完成后的聊天对象变化
                        if issubclass(chat_type, Group):

                            if username in self.data.raw_chats:
                                # 群成员列表更新
                                before = self.data.raw_chats[username]['MemberList']
                                after = raw_dict['MemberList']

                                old, new = diff_usernames(before, after)
                                group_username = raw_dict['UserName']

                                list(map(lambda x: self.deleting_member(Member(self, x, group_username)),
                                         filter(lambda x: x['UserName'] in old, before)))

                                list(map(lambda x: self.new_member(Member(self, x, group_username)),
                                         filter(lambda x: x['UserName'] in new, after)))
                            else:
                                # 新增群聊
                                self.new_group(chat_type(self, raw_dict))

                        elif issubclass(chat_type, Friend) and username not in self.data.raw_chats:
                            # 新增好友
                            self.new_friend(chat_type(self, raw_dict))

                    self.data.raw_chats[username] = raw_dict

    def put_new_messages(self, raw_msg_list):
        """ 将新消息加入消息处理队列 """

        for raw_msg in raw_msg_list:
            # noinspection PyBroadException
            try:
                msg = Message(self, raw_msg)
            except:
                logger.exception('failed to parse message:\n{}'.format(raw_msg))
            else:
                self.new_message(msg)
                self.message_queue.put(msg)

    # [utils]

    def load_response_as_json(self, resp):

        """
        解析 Response 中的 JSON 数据，并同步各

        :param resp: :class:`requests.Response` 对象
        """

        resp.encoding = 'utf-8'
        json_dict = resp.json(object_hook=decode_webwx_json_values)

        skey = json_dict.get('SKey')
        if skey:
            self.data.skey = skey
        sync_key = json_dict.get('SyncKey')
        if sync_key:
            self.data.sync_key = sync_key
        sync_check_key = json_dict.get('SyncCheckKey')
        if sync_check_key:
            self.data.sync_check_key = sync_check_key

        if 'BaseResponse' in json_dict:
            try:
                base_response = json_dict['BaseResponse']
                err_code = base_response['Ret']
                err_msg = base_response['ErrMsg']
            except (KeyError, TypeError):
                logger.error('failed to parse base_response:\n{}'.format(json_dict['BaseResponse']))
            else:
                if err_code != 0:
                    raise ResponseError(self, err_code, err_msg)

        return json_dict

    def get_chats(self, chat_type=None, convert=False):
        """
        从 Core.data.raw_chats 中过滤出指定类型的聊天对象列表

        :param chat_type: 传入基于 :class:`Chat` 的 class，来指定聊天对象类型
        :param convert: 将获得的 dict 转换为指定的类
        """

        chat_list = list(self.data.raw_chats.values())

        if chat_type:
            chat_list = list(filter(
                lambda x: issubclass(get_chat_type(x), chat_type),
                chat_list
            ))

        if convert:

            for i, raw_chat in enumerate(chat_list):
                to_class = get_chat_type(raw_chat) if chat_type in (None, MP) else chat_type
                username = raw_chat['UserName']
                chat_list[i] = to_class(self, username)

            chat_list = Chats(chat_list, self)

        return chat_list

    def get_chat_obj(self, username, group_username=None):
        """
        | 将 username 转化为聊天对象
        | 若为群成员，则需填 group_username
        """

        def fetch_contact(_username):
            if _username not in self.data.raw_chats:
                # Todo: 如果 UserName 为 fmessage 则返回一个昵称为 朋友推荐消息 的聊天对象
                # Todo: 如果 UserName 为 filehelper 则返回一个昵称为 文件传输助手 的聊天对象
                self.batch_get_contact(_username)
            if _username in self.data.raw_chats:
                raw_chat = self.data.raw_chats[_username]
                return get_chat_type(raw_chat)(self, raw_chat)
            else:
                logger.warning('chat not found: {}'.format(username))
                return Chat(self, username)

        if group_username:
            group = fetch_contact(group_username)
            member = group.find(username=username)
            if not member:
                group.update()
                member = group.find(username=username)
            return member
        else:
            return fetch_contact(username)

    def from_cookies(self, name):
        """ 从 cookies 中获取值 """
        return self.session.cookies.get(name)

    # extend
    def update_all_data(self):
        """ 更新数据: raw_chat,  """
        if self.session and self.alive:
            try:
                self.sync(retries=1, delete_when_modify=True)
            except ResponseError:
                logger.info('failed to continue last data sync loop')
            else:
                self.self = Friend(self, self.data.raw_self)

                self.dump()

    def is_living(self):
        """ 判断 WeChat 机器人是否登录有效

        :rtype: bool
        """
        return self.alive is True


def from_js(js, *names):
    """ 从 js 请求中获取指定的一个或多个变量名对应的值，需要时会自动转换为数字 """

    if names:
        def get_value(name):
            found = re.search(r'\b' + re.escape(name) + r'\s*=\s*("|\'|)(.+?)\1\s*;', js)
            if found:
                v = found.group(2)
                if found.group(1):
                    return v
                elif '.' in v:
                    return float(v)
                else:
                    return int(v)

        try:
            ret = smart_map(get_value, names)
        except ValueError:
            raise ValueError('invalid js:\n{}'.format(js))
        if len(ret) == 1:
            return ret[0]
        return ret


def merge_chat_dict(old, new):
    for k, v in new.items():
        if v and v not in (old.get(k), {'Buff': ''}):
            if isinstance(v, dict):
                if 'Buff' in v:
                    old[k] = v['Buff']
                    continue
            old[k] = v


def prompt(content):
    print('* {}'.format(content))



# coding: utf-8
from __future__ import unicode_literals

from wxpy.compatible.utils import force_encoded_string_output


class MessageType(object):
    def __init__(self, name='UNKNOWN', main=None, app=None, sub=None):

        """
        消息类型

        :param name: 友好类型名称
        :param main: 主要类型
        :param app: 应用类型
        :param sub: 子类型
        """

        self.name = name

        self.main = main
        self.app = app
        self.sub = sub

    @force_encoded_string_output
    def __repr__(self):
        return '<MsgType: {} ({}, {}, {})>'.format(
            self.name, self.main, self.app, self.sub)

    def __str__(self):
        return self.name

    def __unicode__(self):
        return self.__str__()

    def __eq__(self, other):
        """
        | 对比两者的各类型 ID, 一方为 None 时不限制
        | 或直接字符串匹配名称
        """

        if isinstance(other, MessageType):
            match = 0
            for type_name in 'main', 'app', 'sub':
                a = getattr(self, type_name)
                b = getattr(other, type_name)
                if None in (a, b):
                    continue
                if a != b:
                    return False
                match += 1
            return match
        else:
            return other == self.name

    def __cmp__(self, other):
        return 0 if self.__eq__(other) else 1

    def __hash__(self):
        return hash((self.name, self.main, self.app))


# 文本
TEXT = MessageType('TEXT', 1, 0, 0)
# 位置
LOCATION = MessageType('LOCATION', 1, 0, 48)
# 图片
IMAGE = MessageType('IMAGE', 3, 0, 0)
# 语音
VOICE = MessageType('VOICE', 34, 0, 0)
# 好友验证
NEW_FRIEND = MessageType('NEW_FRIEND', 37, 0, 0)
# 名片
CARD = MessageType('CARD', 42, 0, 0)
# 视频
VIDEO = MessageType('VIDEO', 43, 0, 0)
# 表情 (不支持商店表情，下载前请先检查 file_size 属性)
STICKER = MessageType('STICKER', 47, 0, 0)
# URL
URL = MessageType('URL', 49, 5, 0)
# 文件
FILE = MessageType('FILE', 49, 6, 0)
# 转账
CASH = MessageType('CASH', 49, 2000, 0)
# 系统提示
NOTICE = MessageType('NOTICE', 10000, 0, 0)
# 撤回提示
RECALLED = MessageType('RECALLED', 10002, 0, 0)

KNOWN_MSG_TYPES = (
    TEXT, LOCATION, IMAGE, VOICE,
    NEW_FRIEND, CARD, VIDEO, STICKER,
    URL, FILE, CASH, NOTICE, RECALLED
)

# 未知
UNKNOWN = MessageType()



# coding: utf-8
from __future__ import unicode_literals

import logging
from collections import Counter

from wxpy.utils import match_attributes, match_name, ensure_one

logger = logging.getLogger(__name__)


class Chats(list):
    """
    多个聊天对象的列表，可用于搜索、更新、统计等功能
    """

    def __init__(self, chat_list=None, source=None):
        if chat_list:
            super(Chats, self).__init__(chat_list)
        self.source = source

    def __add__(self, other):
        return Chats(
            super(Chats, self).__add__(other or list()),
            self.source if self.source == getattr(other, 'source', None) else None
        )

    def _iter_match(self, keywords=None, **attributes):
        def match(_chat):

            if not match_name(_chat, keywords):
                return
            if not match_attributes(_chat, **attributes):
                return
            return True

        for chat in self:
            if match(chat):
                yield chat

    def get(self, keywords=None, **attributes):

        """
        | 找到匹配所设条件的唯一聊天对象
        | 若结果不唯一，或没有找到，会抛出 `ValueError` 异常

        :param keywords:

            | 聊天对象的名称关键词 (空格分隔, 不区分大小写)
            | 用于在 remark_name, display_name, nickname, wxid 中进行搜索

        :param attributes:

            属性键值对，键可以是 sex(性别), province(省份), city(城市) 等，例如::

                bot.chats.get(nickname='游否', city='深圳', sex=MALE)

        :rtype: :class:`Chat`
        """

        found = list(self._iter_match(keywords, **attributes))
        return ensure_one(found)

    def find(self, keywords=None, **attributes):
        """
        | 类似于 any:`Chats.get`，不同之处在于仅返回 **首个匹配的** 聊天对象
        | 也就是说: 即便列表内有多个匹配，也不会抛出异常; 若没有匹配的，则返回 None

        :rtype: :class:`Chat`
        """

        for chat in self._iter_match(keywords, **attributes):
            return chat

    def search(self, keywords=None, **attributes):
        """
        类似于 any:`Chats.get`，不同之处在于会返回 **所有匹配的** 聊天对象

        :rtype: :class:`Chats`
        """

        return Chats(self._iter_match(keywords, **attributes), self.source)

    def update(self):
        """
        更新列表中的所有聊天对象
        """

        cores = dict()
        for chat in self:
            cores.setdefault(chat.core, list()).append(chat)
        for core in cores:
            core.batch_get_contact(cores[core])

    def stats(self, attribs=('sex', 'province', 'city')):
        """
        统计各属性的分布情况

        :param attribs: 需统计的属性列表或元组
        :return: 统计结果
        """

        def attr_stat(objects, attr_name):
            return Counter(list(map(lambda x: getattr(x, attr_name), objects)))

        from wxpy.utils import ensure_list
        attribs = ensure_list(attribs)
        ret = dict()
        for attr in attribs:
            ret[attr] = attr_stat(self, attr)
        return ret

    def stats_text(self, total=True, sex=True, top_provinces=10, top_cities=10):
        """
        简单的统计结果的文本

        :param total: 总体数量
        :param sex: 性别分布
        :param top_provinces: 省份分布
        :param top_cities: 城市分布
        :return: 统计结果文本
        """

        from .group import Group
        from wxpy.api.consts import FEMALE
        from wxpy.api.consts import MALE
        from wxpy.api.core import Core

        def top_n_text(attr, n):
            top_n = list(filter(lambda x: x[0], stats[attr].most_common()))[:n]
            top_n = ['{}: {} ({:.2%})'.format(k, v, v / len(self)) for k, v in top_n]
            return '\n'.join(top_n)

        stats = self.stats()

        text = str()

        if total:
            if self.source:
                if isinstance(self.source, Core):
                    user_title = '微信好友'
                    nickname = self.source.name
                elif isinstance(self.source, Group):
                    user_title = '群成员'
                    nickname = self.source.nickname
                else:
                    raise TypeError('source should be Bot or Group')
                text += '{nickname} 共有 {total} 位{user_title}\n\n'.format(
                    nickname=nickname,
                    total=len(self),
                    user_title=user_title
                )
            else:
                text += '共有 {} 位用户\n\n'.format(len(self))

        if self:
            if sex:
                males = stats['sex'].get(MALE, 0)
                females = stats['sex'].get(FEMALE, 0)

                text += '男性: {males} ({male_rate:.1%})\n女性: {females} ({female_rate:.1%})\n\n'.format(
                    males=males,
                    male_rate=males / len(self),
                    females=females,
                    female_rate=females / len(self),
                )

            if top_provinces:
                text += 'TOP {} 省份\n{}\n\n'.format(
                    top_provinces,
                    top_n_text('province', top_provinces)
                )

            if top_cities:
                text += 'TOP {} 城市\n{}\n\n'.format(
                    top_cities,
                    top_n_text('city', top_cities)
                )

        return text



# coding: utf-8
from __future__ import unicode_literals

import logging

from wxpy.compatible.utils import force_encoded_string_output
from wxpy.utils import repr_message

logger = logging.getLogger(__name__)


class SentMessage(object):
    """
    程序中通过 .send/reply() 系列方法发出的消息
    
    *使用程序发送的消息也将被记录到历史消息 bot.messages 中*
    """

    def __init__(self, **attributes):

        self.core = None

        # 消息的类型
        self.type = None

        # 消息的服务端 ID
        self.id = None

        # 消息的本地 ID (撤回时需要用到)
        self.local_id = None

        # 消息的文本内容
        self.text = None

        # 消息附件的本地路径
        self.path = None

        # 消息的附件 media_id
        self.media_id = None

        # 本地发送时间
        self.create_time = None

        # 接收服务端响应时间
        self.receive_time = None

        self.receiver = None

        for k, v in attributes.items():
            setattr(self, k, v)

    def __hash__(self):
        return hash((SentMessage, self.id))

    @force_encoded_string_output
    def __repr__(self):
        return repr_message(self)

    def __unicode__(self):
        return repr_message(self)

    @property
    def latency(self):
        """
        消息的延迟秒数 (发送时间和响应时间的差值)
        """
        if self.create_time and self.receive_time:
            return (self.receive_time - self.create_time).total_seconds()

    @property
    def chat(self):
        """
        消息所在的聊天会话 (始终为消息的接受者)
        """
        return self.receiver

    @property
    def member(self):
        """
        若在群聊中发送消息，则为群员
        """
        from wxpy import Group

        if isinstance(self.receiver, Group):
            return self.receiver.self

    @property
    def bot(self):
        """
        消息所属的机器人
        """
        return self.receiver.bot

    @property
    def sender(self):
        """
        消息的发送者
        """
        return self.receiver.bot.self

    def recall(self):
        """
        撤回本条消息 (应为 2 分钟内发出的消息)
        """

        logger.info('recalling msg:\n{}'.format(self))

        return self.core.revoke_msg(self.receiver, self.id, self.local_id)



# coding: utf-8
from __future__ import unicode_literals

import inspect
import logging
import re
import threading
import time

from requests.adapters import HTTPAdapter

from wxpy.compatible import PY2

if PY2:
    from future.builtins import str
    # noinspection PyUnresolvedReferences
    import HTMLParser
else:
    import html


def ensure_list(x, except_false=True):
    """
    若传入的对象不为列表，则转化为列表

    :param x: 输入对象
    :param except_false: None, False 等例外，会直接返回原值
    :return: 列表，或 None, False 等
    :rtype: list
    """

    if isinstance(x, (list, tuple)) or (not x and except_false):
        return x
    return [x]


def prepare_keywords(keywords):
    """
    准备关键词
    """

    if not keywords:
        keywords = ''
    if isinstance(keywords, str):
        # noinspection PyTypeChecker
        keywords = re.split(r'\s+', keywords)
    return map(lambda x: x.lower(), keywords)


def match_text(text, keywords):
    """
    判断文本内容中是否包含了所有的关键词 (不区分大小写)

    :param text: 文本内容
    :param keywords: 关键词，可以是空白分割的 str，或是多个精准关键词组成的 list
    :return: 若包含了所有的关键词则为 True，否则为 False
    """

    if not text:
        text = ''
    else:
        text = text.lower()

    keywords = prepare_keywords(keywords)

    for kw in keywords:
        if kw not in text:
            return False
    return True


def match_attributes(obj, **attributes):
    """
    判断对象是否匹配输入的属性条件

    :param obj: 对象
    :param attributes: 属性键值对
    :return: 若匹配则为 True，否则为 False
    """

    for attr, value in attributes.items():
        if getattr(obj, attr, None) != value:
            return False
    return True


def match_name(chat, keywords):
    """
    判断一个 Chat 对象的名称是否包含了所有的关键词 (不区分大小写)

    :param chat: Chat 对象
    :param keywords: 关键词，可以是空白分割的 str，或是多个精准关键词组成的 list
    :return: 若包含了所有的关键词则为 True，否则为 False
    """
    keywords = prepare_keywords(keywords)

    for kw in keywords:
        for attr in 'remark_name', 'display_name', 'nickname', 'wxid':
            if kw in '{0}'.format(getattr(chat, attr, '')).lower():
                break
        else:
            return False
    return True


def smart_map(func, i, *args, **kwargs):
    """
    将单个对象或列表中的每个项传入给定的函数，并返回单个结果或列表结果，类似于 map 函数

    :param func: 传入到的函数
    :param i: 列表或单个对象
    :param args: func 函数所需的 args
    :param kwargs: func 函数所需的 kwargs
    :return: 若传入的为列表，则以列表返回每个结果，反之为单个结果
    """
    if isinstance(i, (list, tuple, set)):
        return list(map(lambda x: func(x, *args, **kwargs), i))
    else:
        return func(i, *args, **kwargs)


def get_raw_dict(chat_or_chats):
    """
    确保将用户转化为带有 UserName 键的用户字典

    :param chat_or_chats: 单个用户，或列表形式的多个用户
    :return: 单个用户字典，或列表形式的多个用户字典
    """

    from wxpy.api.chats import Chat

    def wrap_one(x):
        if isinstance(x, dict):
            return x
        elif isinstance(x, Chat):
            return x.raw
        elif isinstance(x, str):
            return {'UserName': chat_or_chats}
        else:
            if PY2:
                # noinspection PyUnresolvedReferences
                if isinstance(x, unicode):
                    return {'UserName': chat_or_chats}
            raise TypeError('unsupported type: {}'.format(type(x)))

    return smart_map(wrap_one, chat_or_chats)


def get_username(chat_or_chats):
    """
    确保将用户转化为 username 字串

    :param chat_or_chats: 单个用户，或列表形式的多个用户
    :return: 返回单个 username 字串，或列表形式的多个 username 字串
    """

    from wxpy.api.chats import Chat
    from wxpy.api.bot import Bot
    from wxpy.api.core import Core

    def get_one(x):
        if isinstance(x, (Chat, Bot, Core)):
            return x.username
        elif isinstance(x, dict):
            return x['UserName']
        elif isinstance(x, str):
            return x
        else:
            if PY2:
                # noinspection PyUnresolvedReferences
                if isinstance(x, unicode):
                    return x
            raise TypeError('Unsupported type: {}'.format(type(x)))

    return smart_map(get_one, chat_or_chats)


def get_receiver(receiver=None):
    """
    获得作为接收者的聊天对象

    :param receiver:
        * 当为 `None`, `True` 或字符串时，将以该值作为 `cache_path` 参数启动一个新的机器人，并返回该机器人的"文件传输助手"
        * 当为 :class:`机器人 <Bot>` 时，将返回该机器人的"文件传输助手"
        * 当为 :class:`聊天对象 <Chat>` 时，将返回该聊天对象
    :return: 作为接收者的聊天对象
    :rtype: :class:`wxpy.Chat`
    """

    from wxpy.api.chats import Chat
    from wxpy.api.bot import Bot

    if isinstance(receiver, Chat):
        return receiver
    elif isinstance(receiver, Bot):
        return receiver.file_helper
    elif receiver in (None, True) or isinstance(receiver, str):
        return Bot(cache_path=receiver).file_helper
    else:
        raise TypeError('expected Chat, Bot, str, True or None')


def enhance_connection(session, pool_connections=30, pool_maxsize=30, max_retries=30):
    """
    增强 requests.Session 对象的网络连接性能

    :param session: 需增强的 requests.Session 对象
    :param pool_connections: 最大的连接池缓存数量
    :param pool_maxsize: 连接池中的最大连接保存数量
    :param max_retries: 最大的连接重试次数 (仅处理 DNS 查询, socket 连接，以及连接超时)
    """

    for p in 'http', 'https':
        session.mount(
            '{}://'.format(p),
            HTTPAdapter(
                pool_connections=pool_connections,
                pool_maxsize=pool_maxsize,
                max_retries=max_retries,
                pool_block=False
            ))


def repr_message(msg):
    """
    用于 Message 和 SentMessage 对象的 __repr__ 和 __unicode__
    """

    from wxpy.api.chats import Group

    text = (str(msg.text or '')).replace('\n', ' ↩ ')
    text += ' ' if text else ''

    if msg.sender == msg.core.self:
        ret = '↪ {self.receiver.name}'
    elif isinstance(msg.chat, Group) and msg.member and msg.member != msg.receiver:
        ret = '{self.sender.name} >> {self.member.name}'
    else:
        ret = '{self.sender.name}'

    ret += ' : {text}({self.type})'

    return ret.format(self=msg, text=text)


def get_text_without_at_bot(msg):
    """
    获得 Message 对象中的消息内容，并清理 @ 机器人的部分

    :param msg: Message 对象
    :return: 清理 @ 机器人部分后的文本内容
    :rtype: str
    """

    from wxpy.api.chats import Group

    text = msg.text

    if isinstance(msg.chat, Group):
        name = msg.chat.self.name
        text = re.sub(r'\s*@' + re.escape(name) + r'\u2005?\s*', '', text)

    return text


def start_new_thread(target, args=(), kwargs=None, daemon=True, use_caller_name=False):
    """
    启动一个新的进程，需要时自动为进程命名，并返回这个线程
    
    :param target: 调用目标
    :param args: 调用位置参数
    :param kwargs: 调用命名参数
    :param daemon: 作为守护进程
    :param use_caller_name: 为 True 则以调用者为名称，否则以目标为名称

    :return: 新的进程
    :rtype: threading.Thread
    """

    if use_caller_name:
        # 使用调用者的名称
        name = inspect.stack()[1][3]
    else:
        name = target.__name__

    logging.getLogger(
        # 使用外层的 logger
        inspect.currentframe().f_back.f_globals.get('__name__')
    ).debug('new thread: {}'.format(name))
    if PY2:
        _thread = threading.Thread(
            target=target, args=args, kwargs=kwargs,
            name=name)
        _thread.setDaemon(daemon)
    else:
        _thread = threading.Thread(
            target=target, args=args, kwargs=kwargs,
            name=name, daemon=daemon
        )
    _thread.start()

    return _thread


rp_emoji_span = re.compile(r'<span class="emoji emoji([\da-fA-F]+)"></span>?')

# 参考: https://en.wikipedia.org/wiki/Emoji#Unicode_blocks
rp_emoji_code_point = re.compile(r'1f[\da-f]{3}|[2-3][\da-f]{3}|a9|ae')

# Web 微信中的 emoji 是通过预先准备好的图片来渲染的 (为了不同平台中的 emoji 风格统一)
# 然而在这个特殊的渲染方式中，有一部分 emoji 存在错误的匹配
# 所以当需要需要输出 emoji 字符 (而不是图片) 时，需要使用一个映射表来修正
# 这个映射表需要从上面说的 Web 微信预先准备好的图片中去提取，显然这不现实…
# 下面的映射表 `emoji_mismatch_map`，是从 itchat 的代码中复制的，感谢维护者们的付出! 我要羞羞的抄作业了...
# https://github.com/littlecodersh/ItChat/blob/86a3fefc257800092d040f2f7849e781785fafe1/itchat/utils.py#L47
emoji_mismatch_map = {
    '1f63c': '1f601', '1f639': '1f602', '1f63a': '1f603',
    '1f4ab': '1f616', '1f64d': '1f614', '1f63b': '1f60d',
    '1f63d': '1f618', '1f64e': '1f621', '1f63f': '1f622',
}


def emoji_replace_hook(match):
    code = match.group(1)
    if code in emoji_mismatch_map:
        code = emoji_mismatch_map[code]

    emoji = ''
    for cp in rp_emoji_code_point.findall(code):
        emoji += '\\U{:0>8}'.format(cp).encode('utf-8').decode('unicode-escape', 'replace')
    return emoji


def decode_webwx_emoji(text):
    """ 将文本中的 <span/> 标签还原为 emoji """
    return rp_emoji_span.sub(emoji_replace_hook, text)


def decode_webwx_json_values(dct):
    """ 解码 Web 微信返回的 JSON 中的 values """

    for k, v in dct.items():
        if isinstance(v, str):
            v = v.replace('<br/>', '\n')
            v = decode_webwx_emoji(v)
            if PY2:
                # 3.0 < version < 3.3 使用 html.parser.HTMLParser().unescape
                # https://stackoverflow.com/questions/2360598/how-do-i-unescape-html-entities-in-a-string-in-python-3-1
                v = HTMLParser.HTMLParser().unescape(v)
            else:
                v = html.unescape(v)
            dct[k] = v

    return dct


def get_chat_type(raw_dict):
    """ 区分原始聊天对象字典的所属类型，返回聊天对象类 """
    from wxpy.api.chats import Friend, Group, Member, Service, Subscription

    if raw_dict['UserName'].startswith('@@'):
        return Group
    elif raw_dict['VerifyFlag'] >= 24:
        return Service
    elif raw_dict['VerifyFlag'] >= 8:
        return Subscription
    elif raw_dict.get('EncryChatRoomId') or 'MemberStatus' in raw_dict:
        return Member
    else:
        return Friend


def diff_usernames(old_chats, new_chats):
    """
    比较两个聊天对象列表或字典中的 username 的差异

    :param old_chats: 原来的聊天对象列表或字典
    :param new_chats: 新的聊天对象列表或字典
    :return: 一个元组: (old 中特有的, new 中特有的)
    """

    old, new = list(map(
        lambda x: set(map(
            get_username,
            x.keys() if isinstance(x, dict) else x
        )),
        (old_chats, new_chats),
    ))

    return list(old - new), list(new - old)


def chunks(whole, chunk_size):
    lst = ensure_list(whole)
    for i in range(0, len(lst), chunk_size):
        yield lst[i:i + chunk_size]


def new_local_msg_id():
    return str(int(time.time() * 1e7))



from datetime import datetime

from tests.conftest import wait_for_message
from wxpy import *


def sent_message(sent_msg, msg_type, receiver):
    assert isinstance(sent_msg, SentMessage)
    assert sent_msg.type == msg_type
    assert sent_msg.receiver == receiver
    assert sent_msg.bot == receiver.bot
    assert sent_msg.sender == receiver.bot.self
    assert isinstance(sent_msg.receive_time, datetime)
    assert isinstance(sent_msg.create_time, datetime)
    assert sent_msg.create_time < sent_msg.receive_time


class TestMessage:
    def test_text_message(self, group, friend):
        sent_message(group.send('text'), TEXT, group)
        msg = wait_for_message(group, TEXT)
        assert isinstance(msg, Message)
        assert msg.type == TEXT
        assert msg.text == 'Hello!'
        assert not msg.is_at
        assert msg.chat == group
        assert msg.sender == group
        assert msg.receiver == group.self
        assert msg.member == friend
        assert 0 < msg.latency < 30

        group.send('at')
        msg = wait_for_message(group, TEXT)
        assert msg.is_at

    def test_picture_message(self, group, image_path):
        sent = group.send_image(image_path)
        sent_message(sent, PICTURE, group)
        assert sent.path == image_path

    def test_video_message(self, group, video_path):
        sent = group.send_video(video_path)
        sent_message(sent, VIDEO, group)
        assert sent.path == video_path

    def test_raw_message(self, group):
        # 发送名片
        raw_type = 42
        raw_content = '<msg username="{}" nickname="{}"/>'.format('wxpy_bot', 'wxpy 机器人')
        sent_message(group.send_raw_msg(raw_type, raw_content), None, group)

    def test_send(self, friend, file_path, image_path, video_path):
        text_to_send = 'test sending text'
        sent = friend.send(text_to_send)
        sent_message(sent, TEXT, friend)
        assert sent.text == text_to_send

        sent = friend.send('@fil@{}'.format(file_path))
        sent_message(sent, ATTACHMENT, friend)
        assert sent.path == file_path

        sent = friend.send('@img@{}'.format(image_path))
        sent_message(sent, PICTURE, friend)
        assert sent.path == image_path

        sent = friend.send('@vid@{}'.format(video_path))
        sent_message(sent, VIDEO, friend)
        assert sent.path == video_path

        # 发送名片
        raw_type = 42
        raw_content = '<msg username="{}" nickname="{}"/>'.format('wxpy_bot', 'wxpy 机器人')
        uri = '/webwxsendmsg'
        sent = friend.send_raw_msg(raw_type, raw_content)
        sent_message(sent, None, friend)

        assert sent.type is None
        assert sent.raw_type == raw_type
        assert sent.raw_content == raw_content
        assert sent.uri == uri



# coding: utf-8
from __future__ import unicode_literals

from .user import User


class Member(User):
    """
    群聊成员对象
    """

    def __init__(self, core, _raw, group_username):
        super(Member, self).__init__(core, _raw)
        self.group_username = group_username

    @property
    def group(self):
        from .group import Group
        return Group(self.core, self.group_username)

    @property
    def display_name(self):
        """
        在群聊中的显示昵称
        """
        return self._chat.get('DisplayName') or None

    @property
    def name(self):
        return self._chat.get('DisplayName') or self._chat.get('NickName') or None

    def remove(self):
        """
        从群聊中移除该成员
        """
        return self.group.remove(self)

    @property
    def raw(self):

        # 群成员的属性获取来源优先级:
        # 1. Member.raw # display_name, nickname, username 会在这层获取到
        # 2. Data.raw_chats # 如果已经是好友，性别地区签名等其他扩展属性会在这层获取到
        # 3. Data.raw_members # 如果不是好友，其他信息会在这层获取到，并可跨群共享

        # 如此可以:
        # 1. 避免合并属性字典(合并会带来内容准确度问题，例如当群员删除群内显示名称时)
        # 2. 共享已获取到的属性 (适用于当群员为好友，或一人多群)

        # noinspection PyProtectedMember
        self.group._complete_member_details()

        return self.core.data.raw_chats.get(self.username) or self.core.data.raw_members[self.username]



# coding: utf-8
# from __future__ import unicode_literals

import re
import codecs

from setuptools import find_packages, setup

with codecs.open('wxpy/__init__.py', encoding='utf-8') as fp:
    version = re.search(r"__version__\s*=\s*'([\w\-.]+)'", fp.read()).group(1)

with codecs.open('README.rst', encoding='utf-8') as fp:
    readme = fp.read()

setup(
    name='bl-wxpy',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'wxpy = wxpy.utils:shell_entry'
        ]
    },
    install_requires=[
        'requests>=2.4.0',
        'pyqrcode',
        'pypng',
        'future',
    ],
    tests_require=[
        'pytest',
    ],
    url='https://github.com/frkhit/bl-wxpy',
    license='MIT',
    author='frkhit',
    author_email='frkhit@gmail.com',
    description='wxpy 定制版, 基于 https://github.com/youfou/wxpy:new-core 分支(不使用 itChat)',
    long_description=readme,
    keywords=[
        '微信',
        'WeChat',
        'API'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
        'Topic :: Communications :: Chat',
        'Topic :: Utilities',
    ]
)



# coding: utf-8

from .user import User


class MP(User):
    """
    公众号对象
    """
    pass


class Service(MP):
    """
    服务号对象
    """
    pass


class Subscription(MP):
    """
    订阅号对象
    """
    pass



# coding: utf-8
from __future__ import unicode_literals

import logging
import threading

from wxpy.utils import ensure_list, get_username
from .chat import Chat
from .chats import Chats
from .member import Member

logger = logging.getLogger(__name__)


class Group(Chat):
    """
    群聊对象
    """

    def __init__(self, core, _chat):
        super(Group, self).__init__(core, _chat)
        self._complete_lock = threading.Lock()

    @property
    def members(self):
        """
        群聊的成员列表

        :rtype: class:Chats:
        """

        return Chats(map(
            lambda x: Member(self.core, x, self.username),
            self.raw['MemberList']
        ), self)

    def __contains__(self, user):
        username = get_username(user)
        for raw_member in self.raw['MemberList']:
            if username == raw_member.get('UserName'):
                return Member(self.core, raw_member, self.username)

    def __iter__(self):
        for member in self.members:
            yield member

    def __len__(self):
        return len(self.members)

    def __getitem__(self, item):
        return self.members.__getitem__(item)

    @property
    def nickname(self):
        """
        该群聊对象的名称
        """

        return super(Group, self).nickname or '、'.join(map(lambda x: x.name, self.members))

    def get(self, keywords=None, **attributes):
        """
        在群聊中查找群成员，等同于 :any:`Group.members.get(...) <Chats.get>`
        """
        return self.members.get(keywords, **attributes)

    def find(self, keywords=None, **attributes):
        """
        在群聊中查找群成员，等同于 :any:`Group.members.find(...) <Chats.find>`
        """
        return self.members.find(keywords, **attributes)

    def search(self, keywords=None, **attributes):
        """
        在群聊中查找群成员，等同于 :any:`Group.members.search(...) <Chats.search>`
        """
        return self.members.search(keywords, **attributes)

    @property
    def owner(self):
        """
        返回群主对象
        """
        owner_username = self.raw.get('ChatRoomOwner')
        if owner_username:
            for member in self:
                if member.username == owner_username:
                    return member
        elif self.members:
            return self.members[0]

    @property
    def is_owner(self):
        """
        判断所属 bot 是否为群管理员
        """
        return self.raw.get('IsOwner') or self.owner.username == self.core.username

    @property
    def self(self):
        """
        机器人自身 (作为群成员)
        """
        return self.__contains__(self.core.data.raw_self)

    def _complete_member_details(self):
        """
        补全群聊中不完整的群员详细信息
        """

        with self._complete_lock:
            to_complete = list()

            for mb in self.raw['MemberList']:
                if mb['UserName'] not in self.core.data.raw_chats \
                        and mb['UserName'] not in self.core.data.raw_members:
                    to_complete.append({'UserName': mb['UserName'], 'ChatRoomId': self.username})

            if to_complete:
                self.core.batch_get_contact(to_complete)

    # noinspection SpellCheckingInspection
    def add(self, members, use_invitation=False):
        """
        向群聊中加入用户

        :param members: 待加入的用户列表或单个用户
        :param use_invitation: 使用发送邀请的方式
        """

        logger.info('adding {} into {} (use_invitation={}))'.format(members, self, use_invitation))

        if use_invitation:
            func_name, key_name = 'invitemember', 'InviteMemberList'
        else:
            func_name, key_name = 'addmember', 'AddMemberList'

        return self.core.update_chatroom(
            self, func_name,
            {key_name: ','.join(map(get_username, ensure_list(members)))}
        )

    def remove(self, members):
        """
        从群聊中移除用户

        :param members: 待移除的用户列表或单个用户
        """

        logger.info('removing {} from {}'.format(members, self))

        # noinspection SpellCheckingInspection
        return self.core.update_chatroom(
            self, 'delmember',
            {'DelMemberList': ','.join(map(get_username, ensure_list(members)))}
        )

    def rename(self, topic):
        """
        修改群聊名称

        :param topic: 新的名称，超长部分会被截断 (最长32字节)
        """

        encodings = ('gbk', 'utf-8')

        trimmed = False

        for ecd in encodings:
            for length in range(32, 24, -1):
                try:
                    # noinspection PyUnresolvedReferences
                    topic = bytes(topic.encode(ecd))[:length].decode(ecd)
                except (UnicodeEncodeError, UnicodeDecodeError):
                    continue
                else:
                    trimmed = True
                    break
            if trimmed:
                break

        logger.info('renaming group: {} => {}'.format(self.name, topic))

        # noinspection SpellCheckingInspection
        return self.core.update_chatroom(
            self, 'modtopic',
            {'NewTopic': topic}
        )



import re

from wxpy import *


class TestUser:
    def test_friend_attributes(self, friend):
        assert isinstance(friend, Friend)
        assert friend.nickname == 'wxpy 机器人'
        assert friend.wxid in ('wxpy_bot', None)
        assert friend.province == '广东'
        assert friend.city == '深圳'
        assert friend.sex == MALE
        assert friend.signature == '如果没有正确响应，可能正在调试中…'
        assert re.match(r'@[\da-f]{32,}', friend.username)
        assert friend.is_friend

    # def test_add(self, member):
    #     member.add('wxpy tests: test_add')

    def test_accept(self, member):
        # 似乎只要曾经是好友，就可以调用这个方法，达到"找回已删除的好友"的效果
        member.accept()

    def test_remark_name(self, friend, member):
        new_remark_name = '__test__123__'

        for user in friend, member:
            current_remark_name = user.remark_name or ''
            for remark_name in new_remark_name, current_remark_name:
                user.set_remark_name(remark_name)



# coding: utf-8
from __future__ import unicode_literals

import logging
import os
import re
from collections import namedtuple
from contextlib import closing
from datetime import datetime
from xml.etree import ElementTree as ETree

from wxpy.api.messages.message_types import *
from wxpy.compatible import PY2
from wxpy.compatible.utils import force_encoded_string_output
from wxpy.utils import repr_message

if PY2:
    # noinspection PyUnresolvedReferences
    from urllib import urlencode
else:
    from urllib.parse import urlencode

logger = logging.getLogger(__name__)

# 公众号推送中的单篇文章内容 (一次可推送多篇)
# 属性: 标题, 摘要, 文章 URL, 封面图片 URL
Article = namedtuple('Article', ['title', 'summary', 'url', 'cover'])

# 转账信息: 金额, 描述, 交易单号 (转账消息可被重复发送，所以要注意核对单号)
Cash = namedtuple('Cash', ['amount', 'description', 'id'])


class Message(object):
    """
    单条消息对象，包括:
    
    * 来自好友、群聊、好友请求等聊天对象的消息
    * 使用机器人账号在手机微信中发送的消息
    
    | 但 **不包括** 代码中通过 .send/reply() 系列方法发出的消息
    | 此类消息请参见 :class:`SentMessage`
    """

    def __init__(self, core, raw):
        self.core = core
        self.bot = self.core.bot

        self.raw = raw

        self._receive_time = datetime.now()

        self._file_ext = None

        # 将 msg.chat.send* 方法绑定到 msg.reply*，例如 msg.chat.send_img => msg.reply_img
        for method in '', '_image', '_sticker', '_file', '_video':
            setattr(self, 'reply' + method, getattr(self.chat, 'send' + method))

    def __hash__(self):
        return hash((Message, self.id))

    @force_encoded_string_output
    def __repr__(self):
        return repr_message(self)

    def __unicode__(self):
        return repr_message(self)

    # basic

    @property
    def type(self):
        """
        消息的类型，目前可为以下值::
        
            # 文本
            TEXT = 'TEXT'
            # 位置
            LOCATION = 'LOCATION'
            # 图片
            IMAGE = 'IMAGE'
            # 语音
            VOICE = 'VOICE'
            # 好友验证
            NEW_FRIEND = 'NEW_FRIEND'
            # 名片
            CARD = 'CARD'
            # 视频
            VIDEO = 'VIDEO'
            # 表情 (不支持商店表情，下载前请先检查 file_size 属性)
            EMOTICON = 'EMOTICON'
            # URL
            URL = 'SHARE_URL'
            # 文件
            FILE = 'FILE'
            # 转账
            CASH = 'CASH'
            # 系统提示
            NOTICE = 'NOTICE'
            # 撤回提示
            RECALLED = 'RECALLED'
            # 未知
            UNKNOWN = 'UNKNOWN'
        
        :rtype: MessageType
        """

        _type = MessageType(
            main=self.raw.get('MsgType'),
            app=self.raw.get('AppMsgType'),
            sub=self.raw.get('SubMsgType'),
        )

        for t in KNOWN_MSG_TYPES:
            if _type == t:
                _type.name = t.name
                break

        return _type

    @property
    def id(self):
        """
        消息的唯一 ID (通常为大于 0 的 64 位整型)
        """
        return self.raw.get('NewMsgId')

    # content

    @property
    def text(self):
        """
        消息的文本内容
        """

        if self.type in (TEXT, NOTICE):
            return self._content
        elif self.type == LOCATION:
            found = re.search(r'^.+(?=:\n)', self._content)
            if found:
                return found.group()
        elif self.type in (URL, CASH):
            return self._content_xml.findtext('.//title')
        elif self.type == NEW_FRIEND:
            return self._content_xml.get('content')
        elif self.type == RECALLED:
            return self._content_xml.findtext('.//replacemsg')
        elif self.type == FILE:
            return self.file_name
        elif self.type == CARD:
            return self.card.name

    @property
    def _file_url(self):

        """
        | 消息中文件的下载地址 (内部使用)
        | 注意: 该 URL 会验证 cookies, 只能使用登陆所在的 session 进行下载
        | 若需下载，请使用 `get_file()` 方法
        """

        uris = self.core.uris
        tree = self._content_xml

        upper_params = {'MsgID': self.id, 'skey': self.core.data.skey, 'type': 'big'}
        lower_params = {'msgid': self.id, 'skey': self.core.data.skey}

        match = {
            IMAGE: (uris.get_msg_img, upper_params),
            STICKER: (uris.get_msg_img, upper_params),
            VOICE: (uris.get_voice, lower_params),
            VIDEO: (uris.get_video, lower_params),
        }.get(self.type)

        if tree and match:
            return '{}?{}'.format(match[0], urlencode(match[1]))
        elif self.type == FILE:
            return '{}?{}'.format(uris.get_media, urlencode(dict(
                sender=self.raw['FromUserName'],
                mediaid=self.media_id,
                filename=self._content_xml.findtext('.//title'),
                fromuser=self.core.data.raw_self['Uin'],
                pass_ticket=self.core.data.pass_ticket,
                webwx_data_ticket=self.core.from_cookies('webwx_data_ticket')
            )))

    @property
    def file_name(self):
        """
        消息中文件的文件名 (含后缀名)
        """
        if self._content_xml and self.type in (IMAGE, STICKER, VOICE, VIDEO):
            return '{}{}'.format(self.id, self.file_ext)
        elif self.type == FILE:
            return self._content_xml.findtext('.//title')

    @property
    def file_ext(self):
        """ 消息中文件的后缀名，例如 .jpeg, .png, .mp3 """

        if self._content_xml and self.type in (IMAGE, STICKER, VOICE, VIDEO):
            if not self._file_ext:
                with closing(self.core.session.get(self._file_url, stream=True)) as resp:
                    if resp.headers.get('Content-Type'):
                        self._file_ext = '.{}'.format(re.findall(r'\w+', resp.headers['Content-Type'])[-1])
            return self._file_ext
        elif self.type == FILE:
            return os.path.splitext(self.file_name)[1]

    def get_file(self, save_path=None):
        """
        下载图片、视频、语音、附件消息中的文件内容。

        可与 :any:`Message.file_name`, :any:`Message.file_ext` 配合使用。

        :param save_path: 文件的保存路径。若为 None，将直接返回字节数据
        """

        if self._file_url:
            return self.core.download(self._file_url, save_path)

    @property
    def file_size(self):
        """
        消息中文件的体积大小
        """
        match = {
            IMAGE: ('img', 'hdlength'),
            STICKER: ('img', 'hdlength'),
            VOICE: ('voicemsg', 'length'),
            VIDEO: ('videomsg', 'length'),
        }.get(self.type)

        if self._content_xml and match:
            return int(self._content_xml.find(match[0]).get(match[1]))
        elif self.type == FILE:
            return int(self._content_xml.findtext('.//totallen'))

    @property
    def media_id(self):
        """
        文件类消息中的文件资源 ID (但图片视频语音等其他消息中为空)
        """
        return self.raw.get('MediaId')

    # group

    @property
    def is_at(self):
        """
        当消息来自群聊，且被 @ 时，为 True
        """

        from wxpy.api.chats import Group

        if self.type == TEXT and isinstance(self.chat, Group):
            return bool(re.search(r'@' + re.escape(self.chat.self.name) + r'(?:\u2005|\s|$)', self.text))

    # misc

    @property
    def url(self):
        """
        分享类消息中的网页 URL
        """
        return self.raw.get('Url')

    @property
    def articles(self):
        """
        公众号推送中的文章列表 (首篇的 标题/地址 与消息中的 text/url 相同)

        其中，每篇文章均有以下属性:

        * `title`: 标题
        * `summary`: 摘要
        * `url`: 文章 URL
        * `cover`: 封面或缩略图 URL
        """

        from wxpy import MP
        if self.type == URL and isinstance(self.sender, MP):
            tree = ETree.fromstring(self.raw['Content'])
            # noinspection SpellCheckingInspection
            items = tree.findall('.//mmreader/category/item')

            article_list = list()

            for item in items:
                article = Article(
                    title=item.findtext('title'),
                    summary=item.findtext('digest'),
                    url=item.findtext('url'),
                    cover=item.findtext('cover'),
                )
                article_list.append(article)

            return article_list

    @property
    def card(self):
        """
        * 好友请求中的请求用户
        * 名片消息中的推荐用户
        """
        from wxpy.api.chats import User

        if self.type in (CARD, NEW_FRIEND):
            return User(self.core, self.raw.get('RecommendInfo'))

    @property
    def recalled_id(self):
        """
        被撤回消息的消息 ID
        """
        if self.type == RECALLED:
            return int(self._content_xml.findtext('.//msgid'))

    @property
    def cash(self):
        if self.type == CASH:
            tree = self._content_xml.find('.//wcpayinfo')
            return Cash(
                amount=float(re.search(r'\d+\.\d+', tree.findtext('feedesc')).group()),
                description=tree.findtext('pay_memo'),
                id=tree.findtext('transcationid'),
            )

    @property
    def img_height(self):
        """
        图片高度
        """
        return self.raw.get('ImgHeight')

    @property
    def img_width(self):
        """
        图片宽度
        """
        return self.raw.get('ImgWidth')

    @property
    def play_length(self):
        """
        视频长度
        """
        return self.raw.get('PlayLength')

    @property
    def voice_length(self):
        """
        语音长度
        """
        return self.raw.get('VoiceLength')

    # time

    @property
    def create_time(self):
        """
        服务端发送时间
        """
        # noinspection PyBroadException
        try:
            return datetime.fromtimestamp(self.raw.get('CreateTime'))
        except:
            pass

    @property
    def receive_time(self):
        """
        本地接收时间
        """
        return self._receive_time

    @property
    def latency(self):
        """
        消息的延迟秒数 (发送时间和接收时间的差值)
        """
        create_time = self.create_time
        if create_time:
            return (self.receive_time - create_time).total_seconds()

    @property
    def location(self):
        """
        位置消息中的地理位置信息
        """
        try:
            ret = ETree.fromstring(self.raw['OriContent']).find('location').attrib
            try:
                ret['x'] = float(ret['x'])
                ret['y'] = float(ret['y'])
                ret['scale'] = int(ret['scale'])
                ret['maptype'] = int(ret['maptype'])
            except (KeyError, ValueError):
                pass
            return ret
        except (TypeError, KeyError, ValueError, ETree.ParseError):
            pass

    # chats

    @property
    def chat(self):
        """
        消息所在的聊天会话，即:

        * 对于自己发送的消息，为消息的接收者
        * 对于别人发送的消息，为消息的发送者
        
        :rtype: :class:`wxpy.User`, :class:`wxpy.Group`
        """

        if self.raw.get('FromUserName') == self.core.username:
            return self.receiver
        else:
            return self.sender

    @property
    def sender(self):
        """
        消息的发送者
        
        :rtype: :class:`wxpy.User`, :class:`wxpy.Group`
        """

        return self.core.get_chat_obj(self.raw.get('FromUserName'))

    @property
    def receiver(self):
        """
        消息的接收者
        
        :rtype: :class:`wxpy.User`, :class:`wxpy.Group`
        """

        return self.core.get_chat_obj(self.raw.get('ToUserName'))

    @property
    def member(self):
        """
        * 若消息来自群聊，则此属性为消息的实际发送人(具体的群成员)
        * 若消息来自其他聊天对象(非群聊)，则此属性为 None
        
        :rtype: NoneType, :class:`wxpy.Member`
        """

        from wxpy.api.chats import Group

        if isinstance(self.chat, Group):
            found = re.search(r'^(@[\da-f]+):\n', self.raw['Content'])
            if found:
                return self.core.get_chat_obj(found.group(1), self.chat.username)
            elif self.type != NOTICE:
                return self.chat.self

    @property
    def _content(self):
        """ Content 字段中去除群员 username 后剩余的部分 """

        from wxpy.api.chats import Group

        if isinstance(self.sender, Group):
            content = re.sub(r'^@[\da-f]+:\n', '', self.raw['Content'])
            if self.type == VIDEO:
                # 发现当有人在群里发视频时，有时 Content 顶部会多 'xxx:\n' 在第二行 (xxx 是该群员的微信 ID)
                content = re.sub(r'^[\da-zA-Z\-_]+:\n', '', content)
            return content
        else:
            return self.raw['Content']

    @property
    def _content_xml(self):
        """ Content 字段中的 xml 对象 """
        try:
            return ETree.fromstring(self._content)
        except ETree.ParseError:
            pass

    def forward(self, chat, prefix=None, suffix=None):
        """
        将本消息转发给其他聊天对象

        支持以下消息类型
            * 文本 (`TEXT`)
            * 图片 (`IMAGE`)
            * 自定义表情 (`STICKER`)
                * 注: 不支持表情商店中的表情
            * 视频（`VIDEO`)
            * 文件 (`FILE`)
            * 名片 (`CARD`)
            * 语音 (`VOICE`)
                * 注: 会以文件方式发送
            * 分享链接 (`URL`)
                * 注: 会转化为 `标题 + 链接` 形式的文本消息
            * 地图 (`LOCATION`)
                * 注: 会转化为 `位置名称 + 地图链接` 形式的文本消息

        :param Chat chat: 接收转发消息的聊天对象
        :param str prefix: 转发时增加的 **前缀** 文本，原消息为文本时会自动换行
        :param str suffix: 转发时增加的 **后缀** 文本，原消息为文本时会自动换行

        :return: 若该消息支持转发，返回转发后的 :class:`SentMessage` 对象；反之返回 `NotImplemented`

        例如，将公司群中的老板消息转发出来::

            from wxpy import *

            bot = Bot()

            # 定位公司群
            company_group = bot.groups.get('公司微信群')

            # 定位老板
            boss = company_group.get('老板大名')

            # 将老板的消息转发到文件传输助手
            @bot.register(company_group)
            def forward_boss_message(msg):
                if msg.member == boss:
                    msg.forward(bot.file_helper, prefix='老板发言')

            # 阻塞线程
            embed()

        """

        logger.info('{}: forwarding to {}: {}'.format(self.bot, chat, self))

        if self.type in (TEXT, NOTICE):
            return chat.send('{}{}{}'.format(
                '{}\n'.format(prefix) if prefix else '',
                self.text,
                '{}\n'.format(suffix) if suffix else '',
            ))

        elif self.type in (IMAGE, STICKER, VIDEO, FILE):
            pass



# coding: utf-8
from __future__ import unicode_literals

import wxpy


class Data(object):
    version = wxpy.__version__

    def __init__(self):
        """
        :class:`Core` 中的所有状态相关的数据内容，方便转存和载入
        """

        self.cookies = None
        self.uris = None

        self.skey = None
        self.sync_key = None
        self.sync_check_key = None

        self.uin = None
        self.pass_ticket = None
        self.gray_scale = None
        self.invite_start_count = None

        self.raw_self = None
        self.raw_chats = dict()
        self.raw_members = dict()
